from osm2gmns.networkclass.macronet import Node, Network
import osm2gmns.settings as og_settings
from shapely import geometry


def _identifyComplexIntersections(network, int_buffer):
    """
    Identify nodes that belongs to one intersection in real life. Nodes that
    belong to one intersection will be assigned with the same intersection_id.
    Only signalized nodes will be checked.

    The reason why only signalized nodes will be checked is that we use a
    distance-based rule here, and there are many short links in osm. If all
    nodes are checked, some short links will be indentified as intersections
    by mistake.

    Parameters
    ----------
    network : Network
        Network instance
    int_buffer : float
        A threshold to check if two connected nodes belong to one intersection.
        If the length of a link that connects two nodes is shorter than int_buffer,
        these two nodes come from one intersection.

    Returns
    -------
    None
    """

    group_list = []
    group_status = []
    for link_id,link in network.link_dict.items():
        if link.length > int_buffer: continue
        if not (link.from_node.intersection_id is None and link.to_node.intersection_id is None): continue
        if not (link.from_node.ctrl_type == 'signal' and link.to_node.ctrl_type == 'signal'): continue
        group_list.append({link.from_node, link.to_node})
        group_status.append(1)

    number_of_valid_groups = sum(group_status)
    while True:
        for group_no1,group1 in enumerate(group_list):
            if group_status[group_no1] == 0: continue
            for group_no2,group2 in enumerate(group_list):
                if group_status[group_no2] == 0: continue
                if group_no1 == group_no2: continue
                if len(group1.intersection(group2)) > 0:
                    group1.update(group2)
                    group_status[group_no2] = 0

        new_number_of_valid_groups = sum(group_status)
        if number_of_valid_groups == new_number_of_valid_groups:
            break
        else:
            number_of_valid_groups = new_number_of_valid_groups

    max_intersection_id = network.max_intersection_id
    for group_no, group in enumerate(group_list):
        if group_status[group_no] == 0: continue
        for node in group: node.intersection_id = max_intersection_id
        max_intersection_id += 1
    network.max_intersection_id = max_intersection_id


def consolidateComplexIntersections(network, auto_identify=False, int_buffer=og_settings.default_int_buffer):
    """
    Consolidate each complex intersection that are original represented by multiple nodes in osm into one node. Nodes
    with the same intersection_id will be consolidated into one node. intersection_id of nodes can be obtained in two ways.
    1) set the argument auto_identify as True, then osm2gmns will automatically identify complex intersections and assign
    intersection_id for corresponding nodes; 2) user can assign intersection_id to nodes manually in network csv files (node.csv), and
    load the network using function loadNetFromCSV provided by osm2gmns. Note that, if the second approach is adopted, one
    should make sure the argument auto_identify is set as False to avoid intersection_id overwritten by osm2gmns.

    Rules used in osm2gmns to identify if two nodes belong to a complex intersection: 1) ctrl_type of the two nodes must be signal;
    2) there is a link connecting these two nodes, and the length of the link is shorter than or equal to the argument int_buffer

    Parameters
    ----------
    network: Network
        osm2gmns Network object
    auto_identify: bool
        if automatically identify complex intersections using built-in methods in osm2gmns. nodes that belong to a complex
        intersection will be assigned with the same intersection_id
    int_buffer: float
        the threshold used to check if two nodes belong to one complex intersection. the unit is meter

    Returns
    -------
    None

    """

    if auto_identify:
        _identifyComplexIntersections(network, int_buffer)

    if og_settings.verbose:
        print('Consolidating Complex Intersections')

    node_group_dict = {}
    node_group_ctrl_type_dict = {}
    for node_id, node in network.node_dict.items():
        if node.intersection_id is not None:
            if node.intersection_id in node_group_dict.keys():
                node_group_dict[node.intersection_id].append(node)
            else:
                node_group_dict[node.intersection_id] = [node]
                node_group_ctrl_type_dict[node.intersection_id] = False
            if node.ctrl_type == 'signal':
                node_group_ctrl_type_dict[node.intersection_id] = True

    removal_node_set = set()
    removal_link_set = set()

    for intersection_id, node_group in node_group_dict.items():
        if len(node_group) < 2:
            continue

        new_node = Node(network.max_node_id)
        new_node.intersection_id = intersection_id
        if node_group_ctrl_type_dict[intersection_id]:
            new_node.ctrl_type = 'signal'
        osm_node_id_list = []
        x_coord_sum, y_coord_sum = 0.0, 0.0
        x_coord_xy_sum, y_coord_xy_sum = 0.0, 0.0

        for node in node_group:
            # node.valid = False
            removal_node_set.add(node)
            osm_node_id_list.append(node.osm_node_id)
            x_coord_sum += node.geometry.x
            y_coord_sum += node.geometry.y
            x_coord_xy_sum += node.geometry_xy.x
            y_coord_xy_sum += node.geometry_xy.y

            for link in node.incoming_link_list:
                if link.from_node in node_group:
                    # link.valid = False
                    removal_link_set.add(link)
                else:
                    link.to_node = new_node
                    new_node.incoming_link_list.append(link)
            for link in node.outgoing_link_list:
                if link.to_node in node_group:
                    # link.valid = False
                    removal_link_set.add(link)
                else:
                    link.from_node = new_node
                    new_node.outgoing_link_list.append(link)

            new_node.osm_highway = node.osm_highway

        new_node.osm_node_id = '_'.join(osm_node_id_list)
        x_coord_ave = round(x_coord_sum / len(node_group), og_settings.lonlat_coord_precision)
        y_coord_ave = round(y_coord_sum / len(node_group), og_settings.lonlat_coord_precision)
        new_node.geometry = geometry.Point(x_coord_ave, y_coord_ave)
        x_coord_xy_ave = round(x_coord_xy_sum / len(node_group), og_settings.local_coord_precision)
        y_coord_xy_ave = round(y_coord_xy_sum / len(node_group), og_settings.local_coord_precision)
        new_node.geometry_xy = geometry.Point(x_coord_xy_ave, y_coord_xy_ave)

        network.node_dict[new_node.node_id] = new_node
        network.max_node_id += 1


    for node in removal_node_set: del network.node_dict[node.node_id]
    for link in removal_link_set: del network.link_dict[link.link_id]


# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 12:05
# @desc         [script description]


from osm2gmns.networkclass.macronet import Network
from osm2gmns.multiresolutionnet.netgen import NetGenerator
from osm2gmns.movement.generate_movements_old import generateMovementsForOneNode
from osm2gmns.utils.util_geo import offsetLine
import osm2gmns.settings as og_settings
from shapely import geometry
from shapely.ops import substring


# todo: connector and centroid

_length_of_short_cut = 0.1
_length_of_cut = {0: 2.0, 1: 8.0, 2: 12.0, 3: 14.0, 4: 16.0, 5: 18.0, 6: 20, 7:22, 8:24}  # e.g. 2:8.0 cut 8 meters if the original macro link has 2 lanes, etc
for i_ in range(9,100): _length_of_cut[i_] = 25
_minimum_cutted_length = 2.0


def _checkMovementLinkNecessity(node_dict):
    for node_id, node in node_dict.items():
        if node.ctrl_type == 'signal': continue

        if len(node.incoming_link_list) == 1 and len(node.outgoing_link_list) >= 1:
            # one imcoming link
            ob_link_set = set()
            multiple_connection = False
            for movement in node.movement_list:
                if movement.ob_link in ob_link_set:
                    multiple_connection = True
                    break
                else:
                    ob_link_set.add(movement.ob_link)
            if multiple_connection: continue

            node.movement_link_needed = False
            ib_link = node.incoming_link_list[0]
            ib_link.downstream_short_cut = True
            ib_link.downstream_is_target = True
            for ob_link in node.outgoing_link_list:
                ob_link.upstream_short_cut = True

        elif len(node.outgoing_link_list) == 1 and len(node.incoming_link_list) >= 1:
            # one outgoing link
            ib_link_set = set()
            multiple_connection = False
            for movement in node.movement_list:
                if movement.ib_link in ib_link_set:
                    multiple_connection = True
                    break
                else:
                    ib_link_set.add(movement.ib_link)
            if multiple_connection: continue

            node.movement_link_needed = False
            ob_link = node.outgoing_link_list[0]
            ob_link.upstream_short_cut = True
            ob_link.upstream_is_target = True
            for ib_link in node.incoming_link_list:
                ib_link.downstream_short_cut = True


def _checkLinkLanes(link_dict):
    for link_id, link in link_dict.items():
        if link.lanes is None:
            print(f'WARNING: lanes information is missing on link {link_id}, default value 1 is used')
            link.lanes_list = [1]


def _linkLaneListFromSegment(link_dict):
    resolution = 5.0
    for link_id, link in link_dict.items():
        lanes_change_point_list_temp = [0, link.length]

        if link.length <= resolution:
            link.lanes_change_point_list = lanes_change_point_list_temp.copy()
        else:
            for segment in link.segment_list:
                if segment.ref_node is link.from_node:      # todo: support link.to_node
                    lanes_change_point_list_temp.append(max(0.0, segment.start_lr))
                    lanes_change_point_list_temp.append(min(link.length, segment.end_lr))

            while lanes_change_point_list_temp:
                target_point = lanes_change_point_list_temp[0]
                remove_list = []
                for item in lanes_change_point_list_temp:
                    if target_point - resolution <= item <= target_point + resolution:
                        remove_list.append(item)
                link.lanes_change_point_list.append(target_point)
                for item in remove_list: lanes_change_point_list_temp.remove(item)
            link.lanes_change_point_list.sort()

        lanes = link.lanes
        link.lanes_list = []
        for i in range(len(link.lanes_change_point_list) - 1):
            link.lanes_list.append(lanes)
            link.lanes_change_list.append([0, 0])
            from_point = link.lanes_change_point_list[i]
            to_point = link.lanes_change_point_list[i+1]
            for segment in link.segment_list:
                length_of_overlapping = min(to_point, segment.end_lr) - max(from_point, segment.start_lr)
                if length_of_overlapping >= resolution:
                    link.lanes_list[-1] += (segment.l_lanes_added + segment.r_lanes_added)
                    link.lanes_change_list[-1][0] += segment.l_lanes_added
                    link.lanes_change_list[-1][1] += segment.r_lanes_added


def _offsetLinkGeometry(link_dict, width_of_lane, GT):
    # check if a link needs offset
    link_offset_dict = {}
    link_ids = tuple(link_dict.keys())
    for link_no_a, link_id_a in enumerate(link_ids):
        link_a = link_dict[link_id_a]
        if link_a in link_offset_dict.keys():
            continue
        geometry_xy_a_r = geometry.LineString(list(link_a.geometry_xy.coords)[::-1])
        reversed_link_found = False
        for link_id_b in link_ids[link_no_a+1:]:
            link_b = link_dict[link_id_b]
            if geometry_xy_a_r.equals_exact(link_b.geometry_xy, tolerance=0.1):
                reversed_link_found = True
                link_offset_dict[link_a] = True
                link_offset_dict[link_b] = True
                break
        if not reversed_link_found:
            link_offset_dict[link_a] = False

    # offset links
    for link, need_offset in link_offset_dict.items():
        if need_offset:
            offset_distance = (link.max_lanes / 2 + 0.5) * width_of_lane
            geometry_xy_offset_ = link.geometry_xy.parallel_offset(distance=offset_distance, side='right', join_style=2)
            if isinstance(geometry_xy_offset_, geometry.MultiLineString):
                link.geometry_xy_offset = offsetLine(link.geometry_xy, offset_distance)
            else:
                link.geometry_xy_offset = geometry.LineString(list(geometry_xy_offset_.coords)[::-1])
            link.geometry_offset = GT.geo_to_latlon(link.geometry_xy_offset)
        else:
            link.geometry_offset = link.geometry
            link.geometry_xy_offset = link.geometry_xy

    # update breakpoints due to the change of geometry
    for link_id, link in link_dict.items():
        link.lanes_change_point_list = [item / link.length * link.length_offset for item in link.lanes_change_point_list]


def _linkLengthToCut(link):
    downstream_max_cut = max(_length_of_short_cut, link.lanes_change_point_list[-1] - link.lanes_change_point_list[-2] - 3)

    if link.upstream_short_cut and link.downstream_short_cut:
        total_length_needed = 2 * _length_of_short_cut + _minimum_cutted_length
        if link.length_offset > total_length_needed:
            link.length_of_cut_upstream = _length_of_short_cut
            link.length_of_cut_downstream = _length_of_short_cut
        else:
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_short_cut
            link.length_of_cut_downstream = link.length_offset / total_length_needed * _length_of_short_cut

    elif link.upstream_short_cut:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > min(downstream_max_cut, _length_of_cut[i]) + _length_of_short_cut + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = _length_of_short_cut
            link.length_of_cut_downstream = min(downstream_max_cut, _length_of_cut[ii])
        else:
            downstream_needed = min(downstream_max_cut, _length_of_cut[0])
            total_length_needed = downstream_needed + _length_of_short_cut + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_short_cut
            link.length_of_cut_downstream = link.length_offset / total_length_needed * downstream_needed

    elif link.downstream_short_cut:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > _length_of_cut[i] + _length_of_short_cut + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = _length_of_cut[ii]
            link.length_of_cut_downstream = _length_of_short_cut
        else:
            total_length_needed = _length_of_cut[0] + _length_of_short_cut + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_cut[0]
            link.length_of_cut_downstream = link.length_offset / total_length_needed * _length_of_short_cut

    else:
        length_found = False
        ii = 0
        for i in range(link.lanes_list[-1], -1, -1):
            if link.length_offset > _length_of_cut[i] + min(downstream_max_cut, _length_of_cut[i]) + _minimum_cutted_length:
                ii = i
                length_found = True
                break

        if length_found:
            link.length_of_cut_upstream = _length_of_cut[ii]
            link.length_of_cut_downstream = min(downstream_max_cut, _length_of_cut[ii])
        else:
            downstream_needed = min(downstream_max_cut, _length_of_cut[0])
            total_length_needed = downstream_needed + _length_of_cut[0] + _minimum_cutted_length
            link.length_of_cut_upstream = link.length_offset / total_length_needed * _length_of_cut[0]
            link.length_of_cut_downstream = link.length_offset / total_length_needed * downstream_needed


def _performLinkCut(link, GT):
    link.cutted_lanes_change_point_list = link.lanes_change_point_list.copy()
    link.cutted_lanes_list = link.lanes_list.copy()
    link.cutted_lanes_change_list = link.lanes_change_list.copy()

    for i in range(1, len(link.cutted_lanes_change_point_list)):
        if link.cutted_lanes_change_point_list[i] > link.length_of_cut_upstream: break
    del link.cutted_lanes_change_point_list[0:i]
    link.cutted_lanes_change_point_list.insert(0, link.length_of_cut_upstream)
    del link.cutted_lanes_list[0:i-1]
    del link.cutted_lanes_change_list[0:i-1]

    for i in range(len(link.cutted_lanes_change_point_list)-2, -1, -1):
        if link.length_offset - link.cutted_lanes_change_point_list[i] > link.length_of_cut_downstream: break
    del link.cutted_lanes_change_point_list[i+1:]
    link.cutted_lanes_change_point_list.append(link.length_offset - link.length_of_cut_downstream)
    del link.cutted_lanes_list[i+1:]
    del link.cutted_lanes_change_list[i+1:]

    # self.cutted_number_of_sections = len(self.cutted_number_of_lanes_list)

    for i in range(len(link.cutted_lanes_list)):
        start_position = link.cutted_lanes_change_point_list[i]
        end_position = link.cutted_lanes_change_point_list[i+1]

        segment_geometry_xy = substring(link.geometry_xy_offset, start_dist=start_position, end_dist=end_position)
        link.cutted_geometry_xy_list.append(segment_geometry_xy)
        link.cutted_geometry_list.append(GT.geo_to_latlon(segment_geometry_xy))


def _cutMacroLinks(link_dict, GT):
    # cut macro links to stop lines to enable movement links generation
    for link_id, link in link_dict.items():
        _linkLengthToCut(link)
        _performLinkCut(link, GT)


def _autoGenerateMovements(macronet):

    max_movement_id = macronet.max_movement_id
    for macronode_id, macronode in macronet.node_dict.items():
        if macronode.movement_list:
            continue
        generateMovementsForOneNode(macronode, max_movement_id, macronet.GT)
        max_movement_id += len(macronode.movement_list)

    macronet.max_movement_id = max_movement_id


def buildMultiResolutionNets(macronet, auto_movement_generation=True, exclusive_bike_walk_lanes=True, connector_type=None,
                             width_of_lane=3.5, length_of_cell=7.0):
    """
    Build the corresponding mesoscopic and microscopic networks for a source (macroscopic) network

    Parameters
    ----------
    macronet: Network
        a soucrce osm2gmns Network object
    auto_movement_generation: bool
        automatically generate movements for intersections without movement information by calling function generateMovements
        in osm2gmns. if auto_movement_generation is set as False, movements at intersections without movement information will not be generated
    exclusive_bike_walk_lanes: bool
        build exclusive lanes for bike and walk
    connector_type: int
        link_type of connectors
    width_of_lane: float
        width of lanes in meter
    length_of_cell: float
        lenght of cells in meter

    Returns
    -------
    None

    """
    if og_settings.verbose:
        print('Building Multiresolution Networks')

    _checkLinkLanes(macronet.link_dict)
    _linkLaneListFromSegment(macronet.link_dict)
    _offsetLinkGeometry(macronet.link_dict, width_of_lane, macronet.GT)

    _checkMovementLinkNecessity(macronet.node_dict)
    _cutMacroLinks(macronet.link_dict, macronet.GT)

    if auto_movement_generation:
        _autoGenerateMovements(macronet)

    net_generator = NetGenerator(macronet, exclusive_bike_walk_lanes, length_of_cell, width_of_lane)
    net_generator.generateNet()
    macronet.mesonet = net_generator.mesonet
    macronet.micronet = net_generator.micronet


from osm2gmns.networkclass.mesonet import MesoNode, MesoLink, MesoNetwork
from osm2gmns.networkclass.micronet import MicroNode, MicroLink, MicroNetwork
from osm2gmns.utils.util_geo import offsetLine
import osm2gmns.settings as og_settings
from shapely import geometry
import sys



class NetGenerator:
    def __init__(self, macronet, exclusive_bike_walk_lanes, length_of_cell, width_of_lane):
        self.macronet = macronet
        # self.exclusive_bike_walk_lanes = exclusive_bike_walk_lanes
        self.exclusive_bike_walk_lanes = False          # todo: update in the next release
        self.length_of_cell = length_of_cell
        self.width_of_lane = width_of_lane

        self.bike_lane_width = 0.5
        self.walk_lane_width = 0.5

        self.mesonet = MesoNetwork()
        self.micronet = MicroNetwork()

        self.number_of_expanded_mesonode = {}       # macronode: number_of_expanded_mesonode


    def getMultimoalUse(self, allowed_uses):
        if self.exclusive_bike_walk_lanes:
            if len(allowed_uses) <= 1:
                return {'mainlane_allowed_uses': allowed_uses, 'extra_bike': False, 'extra_walk': False}
            else:
                allowed_uses_set = set(allowed_uses).union({'auto','bike','walk'})
                if allowed_uses_set == {'auto','bike'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':True, 'extra_walk':False}
                elif allowed_uses_set == {'auto','walk'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':False, 'extra_walk':True}
                elif allowed_uses_set == {'bike','walk'}:
                    return {'mainlane_allowed_uses':['bike'], 'extra_bike':False, 'extra_walk':True}
                elif allowed_uses_set == {'auto','bike','walk'}:
                    return {'mainlane_allowed_uses':['auto'], 'extra_bike':True, 'extra_walk':True}
        else:
            return {'mainlane_allowed_uses':allowed_uses, 'extra_bike':False, 'extra_walk':False}


    def getLaneGeometry(self, original_geometry, lane_offset):
        if lane_offset < -1e-3 or lane_offset > 1e-3:
            lane_geometry_xy_ = original_geometry.parallel_offset(distance=lane_offset, side='right', join_style=2)
            if isinstance(lane_geometry_xy_, geometry.MultiLineString):
                lane_geometry_xy = offsetLine(original_geometry, lane_offset)
            else:
                if lane_offset > 1e-3:
                    lane_geometry_xy = geometry.LineString(list(lane_geometry_xy_.coords)[::-1])
                else:
                    lane_geometry_xy = lane_geometry_xy_

            if lane_geometry_xy.is_empty:
                return self.getLaneGeometry(original_geometry, lane_offset*0.6)

        else:
            lane_geometry_xy = original_geometry

        return lane_geometry_xy


    def createMicroNetForNormalLink(self, link):
        max_micronode_id = self.micronet.max_node_id
        max_microlink_id = self.micronet.max_link_id

        MultimoalUse = self.getMultimoalUse(link.allowed_uses)
        mainlane_allowed_uses, extra_bike, extra_walk = MultimoalUse['mainlane_allowed_uses'], MultimoalUse['extra_bike'], MultimoalUse['extra_walk']

        # create micronodes on each mesolink
        for mesolink in link.mesolink_list:
            original_number_of_lanes = mesolink.macrolink.lanes
            lane_changes_left = mesolink.lanes_change[0]
            num_of_lane_offset_between_left_most_and_central = -1 * (original_number_of_lanes / 2 - 0.5 + lane_changes_left)

            lane_geometry_xy_list = []
            extra_bike_geometry_xy, extra_walk_geometry_xy = None, None
            lane_offset = 0
            for i in range(mesolink.lanes):
                lane_offset = (num_of_lane_offset_between_left_most_and_central + i) * self.width_of_lane
                lane_geometry_xy_list.append(self.getLaneGeometry(mesolink.geometry_xy, lane_offset))
            if extra_bike and not extra_walk:
                bike_lane_offset = lane_offset + self.bike_lane_width
                extra_bike_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, bike_lane_offset)
            if not extra_bike and extra_walk:
                walk_lane_offset = lane_offset + self.walk_lane_width
                extra_walk_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, walk_lane_offset)
            if extra_bike and extra_walk:
                bike_lane_offset = lane_offset + self.bike_lane_width
                walk_lane_offset = bike_lane_offset + self.walk_lane_width
                extra_bike_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, bike_lane_offset)
                extra_walk_geometry_xy = self.getLaneGeometry(mesolink.geometry_xy, walk_lane_offset)

            number_of_cells = max(1, round(mesolink.length / self.length_of_cell))
            micronode_geometry_xy_list = [[lane_geometry_xy.interpolate(i/number_of_cells, normalized=True) for i in range(number_of_cells+1)] for lane_geometry_xy in lane_geometry_xy_list]
            micronode_geometry_xy_bike = [extra_bike_geometry_xy.interpolate(i / number_of_cells, normalized=True) for i in range(number_of_cells + 1)] if extra_bike_geometry_xy is not None else None
            micronode_geometry_xy_walk = [extra_walk_geometry_xy.interpolate(i / number_of_cells, normalized=True) for i in range(number_of_cells + 1)] if extra_walk_geometry_xy is not None else None

            for i in range(mesolink.lanes):
                micronode_list_lane = []
                for micronode_geometry_xy in micronode_geometry_xy_list[i]:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    micronode.lane_no = i + 1
                    micronode_list_lane.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1
                mesolink.micronode_list.append(micronode_list_lane)

            if extra_bike:
                for micronode_geometry_xy in micronode_geometry_xy_bike:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    # micronode.lane_no = i + 1
                    mesolink.micronode_bike.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1

            if extra_walk:
                for micronode_geometry_xy in micronode_geometry_xy_walk:
                    micronode = MicroNode(max_micronode_id)
                    micronode.geometry_xy = micronode_geometry_xy
                    micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                    micronode.mesolink = mesolink
                    # micronode.lane_no = i + 1
                    mesolink.micronode_walk.append(micronode)
                    self.micronet.node_dict[micronode.node_id] = micronode
                    max_micronode_id += 1

        # micronodes on two ends
        first_mesolink = link.mesolink_list[0]
        for micronode_list_lane in first_mesolink.micronode_list:
            micronode_list_lane[0].is_link_upstream_end_node = True
        if extra_bike: first_mesolink.micronode_bike[0].is_link_upstream_end_node = True
        if extra_walk: first_mesolink.micronode_walk[0].is_link_upstream_end_node = True
        last_mesolink = link.mesolink_list[-1]
        for micronode_list_lane in last_mesolink.micronode_list:
            micronode_list_lane[-1].is_link_downstream_end_node = True
        if extra_bike: last_mesolink.micronode_bike[-1].is_link_downstream_end_node = True
        if extra_walk: last_mesolink.micronode_walk[-1].is_link_downstream_end_node = True

        # micronodes between two adjacent mesolinks
        for i in range(len(link.mesolink_list)-1):
            upstream_mesolink = link.mesolink_list[i]
            downstream_mesolink = link.mesolink_list[i+1]

            up_index_of_left_most_lane_of_original_link = upstream_mesolink.lanes_change[0]
            down_index_of_left_most_lane_of_original_link = downstream_mesolink.lanes_change[0]
            min_left_most_lane_index = min(up_index_of_left_most_lane_of_original_link, down_index_of_left_most_lane_of_original_link)
            up_lane_index_start = up_index_of_left_most_lane_of_original_link - min_left_most_lane_index
            down_lane_index_start = down_index_of_left_most_lane_of_original_link - min_left_most_lane_index

            number_of_connecting_lanes = min(upstream_mesolink.lanes-up_lane_index_start,
                                             downstream_mesolink.lanes-down_lane_index_start)

            for j in range(number_of_connecting_lanes):
                up_lane_index = up_lane_index_start + j
                down_lane_index = down_lane_index_start + j
                up_micronode = upstream_mesolink.micronode_list[up_lane_index][-1]
                down_micronode = downstream_mesolink.micronode_list[down_lane_index][0]
                upstream_mesolink.micronode_list[up_lane_index][-1] = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]

            if extra_bike:
                up_micronode = upstream_mesolink.micronode_bike[-1]
                down_micronode = downstream_mesolink.micronode_bike[0]
                upstream_mesolink.micronode_bike[-1] = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]
            if extra_walk:
                up_micronode = upstream_mesolink.micronode_walk[-1]
                down_micronode = downstream_mesolink.micronode_walk[0]
                upstream_mesolink.micronode_walk = down_micronode
                del self.micronet.node_dict[up_micronode.node_id]

        # create cell
        for mesolink in link.mesolink_list:
            # mainline
            for i in range(mesolink.lanes):
                # travelling
                for j in range(len(mesolink.micronode_list[i])-1):
                    microlink = MicroLink(max_microlink_id)
                    microlink.from_node = mesolink.micronode_list[i][j]
                    microlink.to_node = mesolink.micronode_list[i][j+1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1	            # //1:traveling; 2:changing
                    microlink.allowed_uses = mainlane_allowed_uses
                    self.micronet.link_dict[microlink.link_id] = microlink
                    max_microlink_id += 1
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

                # changing
                if i <= mesolink.lanes - 2:
                    # to left
                    for j in range(len(mesolink.micronode_list[i])-1):
                        microlink = MicroLink(max_microlink_id)
                        microlink.from_node = mesolink.micronode_list[i][j]
                        microlink.to_node = mesolink.micronode_list[i+1][j+1]
                        microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                        microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                        microlink.mesolink = mesolink
                        microlink.cell_type = 2	            # //1:traveling; 2:changing
                        microlink.allowed_uses = mainlane_allowed_uses
                        self.micronet.link_dict[microlink.link_id] = microlink
                        max_microlink_id += 1
                        microlink.from_node.outgoing_link_list.append(microlink)
                        microlink.to_node.incoming_link_list.append(microlink)

                if i >= 1:
                    # to right
                    for j in range(len(mesolink.micronode_list[i])-1):
                        microlink = MicroLink(max_microlink_id)
                        microlink.from_node = mesolink.micronode_list[i][j]
                        microlink.to_node = mesolink.micronode_list[i-1][j+1]
                        microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                        microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                        microlink.mesolink = mesolink
                        microlink.cell_type = 2	            # //1:traveling; 2:changing
                        microlink.allowed_uses = mainlane_allowed_uses
                        self.micronet.link_dict[microlink.link_id] = microlink
                        max_microlink_id += 1
                        microlink.from_node.outgoing_link_list.append(microlink)
                        microlink.to_node.incoming_link_list.append(microlink)

            # bike
            if extra_bike:
                for j in range(len(mesolink.micronode_bike) - 1):
                    microlink = MicroLink(max_microlink_id)
                    microlink.from_node = mesolink.micronode_bike[j]
                    microlink.to_node = mesolink.micronode_bike[j + 1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1  # //1:traveling; 2:changing
                    microlink.allowed_uses = ['bike']
                    self.micronet.link_dict[microlink.link_id] = microlink
                    max_microlink_id += 1
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

            # walk
            if extra_walk:
                for j in range(len(mesolink.micronode_walk) - 1):
                    microlink = MicroLink(max_microlink_id)
                    microlink.from_node = mesolink.micronode_walk[j]
                    microlink.to_node = mesolink.micronode_walk[j + 1]
                    microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                    microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                    microlink.mesolink = mesolink
                    microlink.cell_type = 1  # //1:traveling; 2:changing
                    microlink.allowed_uses = ['walk']
                    self.micronet.link_dict[microlink.link_id] = microlink
                    max_microlink_id += 1
                    microlink.from_node.outgoing_link_list.append(microlink)
                    microlink.to_node.incoming_link_list.append(microlink)

        self.micronet.max_node_id = max_micronode_id
        self.micronet.max_link_id = max_microlink_id


    def createMicroNetForConnector(self, mesolink, ib_mesolink, ib_lane_index_start, ob_mesolink, ob_lane_index_start):
        max_micronode_id = self.micronet.max_node_id
        max_microlink_id = self.micronet.max_link_id

        for i in range(mesolink.lanes):
            start_micronode = ib_mesolink.micronode_list[ib_lane_index_start+i][-1]
            end_micronode = ob_mesolink.micronode_list[ob_lane_index_start+i][0]
            lane_geometry_xy = geometry.LineString([start_micronode.geometry_xy, end_micronode.geometry_xy])

            number_of_cells = max(1, round(lane_geometry_xy.length / self.length_of_cell))
            micronode_geometry_xy_list = [lane_geometry_xy.interpolate(i/number_of_cells, normalized=True) for i in range(1,number_of_cells)]

            mesolink.micronode_list.append([])
            mesolink.microlink_list.append([])
            last_micronode = start_micronode

            first_movement_cell = True

            for micronode_geometry_xy in micronode_geometry_xy_list:
                micronode = MicroNode(max_micronode_id)
                micronode.geometry_xy = micronode_geometry_xy
                micronode.geometry = self.macronet.GT.geo_to_latlon(micronode_geometry_xy)
                micronode.mesolink = mesolink
                micronode.lane_no = i + 1
                mesolink.micronode_list[-1].append(micronode)
                self.micronet.node_dict[micronode.node_id] = micronode
                max_micronode_id += 1

                microlink = MicroLink(max_microlink_id)
                microlink.from_node = last_micronode
                microlink.to_node = micronode
                microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
                microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
                microlink.mesolink = mesolink
                microlink.cell_type = 1	            # //1:traveling; 2:changing

                if first_movement_cell:
                    microlink.is_first_movement_cell = True
                    first_movement_cell = False

                mesolink.microlink_list[-1].append(microlink)
                self.micronet.link_dict[microlink.link_id] = microlink
                max_microlink_id += 1
                microlink.from_node.outgoing_link_list.append(microlink)
                microlink.to_node.incoming_link_list.append(microlink)

                last_micronode = micronode

            microlink = MicroLink(max_microlink_id)
            microlink.from_node = last_micronode
            microlink.to_node = end_micronode
            microlink.geometry = geometry.LineString([microlink.from_node.geometry, microlink.to_node.geometry])
            microlink.geometry_xy = geometry.LineString([microlink.from_node.geometry_xy, microlink.to_node.geometry_xy])
            microlink.mesolink = mesolink
            microlink.cell_type = 1	            # //1:traveling; 2:changing

            if first_movement_cell:
                microlink.is_first_movement_cell = True

            mesolink.microlink_list[-1].append(microlink)
            self.micronet.link_dict[microlink.link_id] = microlink
            max_microlink_id += 1
            microlink.from_node.outgoing_link_list.append(microlink)
            microlink.to_node.incoming_link_list.append(microlink)

        self.micronet.max_node_id = max_micronode_id
        self.micronet.max_link_id = max_microlink_id


    def createMesoNodeForCentriod(self):
        # create new meso nodes for centroids
        for node_id, node in self.macronet.node_dict.items():
            if node.is_centroid:

                if node not in self.number_of_expanded_mesonode.keys():
                    self.number_of_expanded_mesonode[node] = 0
                number_of_expanded_mesonode = self.number_of_expanded_mesonode[node]
                self.number_of_expanded_mesonode[node] += 1

                mesonode = MesoNode(node.node_id * 100 + number_of_expanded_mesonode)
                mesonode.geometry = node.geometry
                mesonode.geometry_xy = node.geometry_xy
                mesonode.macronode = node
                node.centroid_mesonode = mesonode
                self.mesonet.node_dict[mesonode.node_id] = mesonode


    def createNormalLinks(self):
        if og_settings.verbose:
            print('  generating normal meso links...')

        max_mesolink_id = self.mesonet.max_link_id

        for link_id, link in self.macronet.link_dict.items():
            macro_from_node = link.from_node
            if macro_from_node.is_centroid:
                upstream_node = macro_from_node.centroid_meso_node
            else:
                if macro_from_node not in self.number_of_expanded_mesonode.keys():
                    self.number_of_expanded_mesonode[macro_from_node] = 0
                number_of_expanded_mesonode = self.number_of_expanded_mesonode[macro_from_node]
                self.number_of_expanded_mesonode[macro_from_node] += 1

                upstream_node = MesoNode(macro_from_node.node_id * 100 + number_of_expanded_mesonode)
                upstream_node.geometry = geometry.Point(link.cutted_geometry_list[0].coords[0])
                upstream_node.geometry_xy = geometry.Point(link.cutted_geometry_xy_list[0].coords[0])
                upstream_node.macronode = macro_from_node

                self.mesonet.node_dict[upstream_node.node_id] = upstream_node

            cutted_number_of_segments = len(link.cutted_lanes_list)
            macro_to_node = link.to_node
            for section_no in range(cutted_number_of_segments):
                if macro_to_node.is_centroid and section_no == cutted_number_of_segments - 1:
                    downstream_node = macro_to_node.centroid_meso_node
                else:
                    if macro_to_node not in self.number_of_expanded_mesonode.keys():
                        self.number_of_expanded_mesonode[macro_to_node] = 0
                    number_of_expanded_mesonode = self.number_of_expanded_mesonode[macro_to_node]
                    self.number_of_expanded_mesonode[macro_to_node] += 1

                    downstream_node = MesoNode(macro_to_node.node_id * 100 + number_of_expanded_mesonode)
                    downstream_node.geometry = geometry.Point(link.cutted_geometry_list[section_no].coords[-1])
                    downstream_node.geometry_xy = geometry.Point(link.cutted_geometry_xy_list[section_no].coords[-1])
                    if section_no == cutted_number_of_segments - 1:
                        downstream_node.macronode = macro_to_node
                    else:
                        downstream_node.macrolink = link

                    self.mesonet.node_dict[downstream_node.node_id] = downstream_node

                mesolink = MesoLink(max_mesolink_id)
                mesolink.from_node = upstream_node
                mesolink.to_node = downstream_node
                mesolink.lanes = link.cutted_lanes_list[section_no]
                mesolink.lanes_change = link.cutted_lanes_change_list[section_no]
                mesolink.geometry = link.cutted_geometry_list[section_no]
                mesolink.geometry_xy = link.cutted_geometry_xy_list[section_no]
                mesolink.macrolink = link

                link.mesolink_list.append(mesolink)
                upstream_node.outgoing_link_list.append(mesolink)
                downstream_node.incoming_link_list.append(mesolink)

                self.mesonet.link_dict[mesolink.link_id] = mesolink
                max_mesolink_id += 1
                upstream_node = downstream_node

            self.createMicroNetForNormalLink(link)

        self.mesonet.max_link_id = max_mesolink_id


    def connectMesoLinksMVMT(self):
        if og_settings.verbose:
            print('  generating movement meso links...')

        max_mesolink_id = self.mesonet.max_link_id

        for macronode_id, macronode in self.macronet.node_dict.items():
            for mvmt in macronode.movement_list:
                ib_link, ob_link = mvmt.ib_link, mvmt.ob_link
                ib_lane_list = [lane_no for lane_no in range(int(mvmt.start_ib_lane), int(mvmt.end_ib_lane + 1))]
                ob_lane_list = [lane_no for lane_no in range(int(mvmt.start_ob_lane), int(mvmt.end_ob_lane + 1))]

                ib_mesolink = ib_link.mesolink_list[-1]
                ob_mesolink = ob_link.mesolink_list[0]

                if len(ib_lane_list) != len(ob_lane_list):
                    print('  warning: number of inbound lanes and outbound lanes at movement {} is not consistent, movement info will be discarded'.format(mvmt.movement_id))
                    continue
                if (0 in ib_lane_list) or (0 in ob_lane_list):
                    print('  warning: lane number 0 detected at movement {}, movement info will be discarded'.format(mvmt.movement_id))
                    continue
                number_of_lanes = len(ib_lane_list)

                # lane index starts from 0
                ib_lane_index_start = ib_mesolink.lanes_change[0] + ib_lane_list[0] if ib_lane_list[0] < 0 else ib_mesolink.lanes_change[0] + ib_lane_list[0] - 1
                ib_lane_index_end = ib_mesolink.lanes_change[0] + ib_lane_list[-1] if ib_lane_list[-1] < 0 else ib_mesolink.lanes_change[0] + ib_lane_list[-1] - 1
                ob_lane_index_start = ob_mesolink.lanes_change[0] + ob_lane_list[0] if ob_lane_list[0] < 0 else ob_mesolink.lanes_change[0] + ob_lane_list[0] - 1
                ob_lane_index_end = ob_mesolink.lanes_change[0] + ob_lane_list[-1] if ob_lane_list[-1] < 0 else ob_mesolink.lanes_change[0] + ob_lane_list[-1] - 1

                if (ib_lane_index_start < 0) or (ob_lane_index_start < 0) or (ib_lane_index_end > ib_mesolink.lanes - 1) or (ob_lane_index_end > ob_mesolink.lanes - 1):
                    print('  warning: inbound or outbound lane info of movement {} is not consistent with what inbound or outbound link has,'
                        ' movement info will be discarded'.format(mvmt.movement_id))
                    continue

                if macronode.movement_link_needed:
                    mesolink = MesoLink(max_mesolink_id)
                    mesolink.from_node = ib_mesolink.to_node
                    mesolink.to_node = ob_mesolink.from_node
                    mesolink.lanes = number_of_lanes
                    mesolink.isconnector = True
                    mesolink.movement = mvmt
                    mesolink.macronode = macronode

                    mesolink.geometry = geometry.LineString([ib_mesolink.geometry.coords[-1], ob_mesolink.geometry.coords[0]])
                    mesolink.geometry_xy = geometry.LineString([ib_mesolink.geometry_xy.coords[-1], ob_mesolink.geometry_xy.coords[0]])

                    self.mesonet.link_dict[mesolink.link_id] = mesolink
                    max_mesolink_id += 1

                    mesolink.from_node.outgoing_link_list.append(mesolink)
                    mesolink.to_node.incoming_link_list.append(mesolink)

                    self.createMicroNetForConnector(mesolink, ib_mesolink, ib_lane_index_start, ob_mesolink, ob_lane_index_start)
                else:
                    if ib_link.downstream_is_target and not ob_link.upstream_is_target:
                        # remove incoming micro nodes and links of ob_mesolink, then connect to ib_mesolink
                        ib_mesolink_to_node = ib_mesolink.to_node
                        ob_mesolink_from_node = ob_mesolink.from_node
                        ob_mesolink.from_node = ib_mesolink_to_node
                        ob_mesolink.geometry = geometry.LineString([ib_mesolink.geometry.coords[-1]] + ob_mesolink.geometry.coords[1:])
                        ob_mesolink.geometry_xy = geometry.LineString([ib_mesolink.geometry_xy.coords[-1]] + ob_mesolink.geometry_xy.coords[1:])
                        del self.mesonet.node_dict[ob_mesolink_from_node.node_id]

                        for i in range(number_of_lanes):
                            ib_lane_index = ib_lane_index_start + i
                            ob_lane_index = ob_lane_index_start + i
                            ib_mesolink_outgoing_micro_node = ib_mesolink.micronode_list[ib_lane_index][-1]
                            ob_mesolink_incoming_micro_node = ob_mesolink.micronode_list[ob_lane_index][0]
                            for microlink in ob_mesolink_incoming_micro_node.outgoing_link_list:
                                microlink.from_node = ib_mesolink_outgoing_micro_node
                            del self.micronet.node_dict[ob_mesolink_incoming_micro_node.node_id]
                    elif not ib_link.downstream_is_target and ob_link.upstream_is_target:
                        # remove outgoing micro nodes and links of ib_mesolink, then connect to ob_mesolink
                        ib_mesolink_to_node = ib_mesolink.to_node
                        ob_mesolink_from_node = ob_mesolink.from_node
                        ib_mesolink.to_node = ob_mesolink_from_node
                        ib_mesolink.geometry = geometry.LineString(ib_mesolink.geometry.coords[:-1] + [ob_mesolink.geometry.coords[0]])
                        ib_mesolink.geometry_xy = geometry.LineString(ib_mesolink.geometry_xy.coords[:-1] + [ob_mesolink.geometry_xy.coords[0]])
                        del self.mesonet.node_dict[ib_mesolink_to_node.node_id]

                        for i in range(number_of_lanes):
                            ib_lane_index = ib_lane_index_start + i
                            ob_lane_index = ob_lane_index_start + i
                            ib_mesolink_outgoing_micro_node = ib_mesolink.micronode_list[ib_lane_index][-1]
                            ob_mesolink_incoming_micro_node = ob_mesolink.micronode_list[ob_lane_index][0]
                            for microlink in ib_mesolink_outgoing_micro_node.incoming_link_list:
                                microlink.to_node = ob_mesolink_incoming_micro_node
                            del self.micronet.node_dict[ib_mesolink_outgoing_micro_node.node_id]
                    else:
                        sys.exit('Target link defintion error')

        self.mesonet.max_link_id = max_mesolink_id





    def generateNet(self):
        self.createMesoNodeForCentriod()
        self.createNormalLinks()
        self.connectMesoLinksMVMT()




import csv
import os
from osm2gmns.utils.util import *
import osmium
from collections import namedtuple
import multiprocessing as mp


Node = namedtuple('Node', ('id', 'tags', 'lonlat'))
Way = namedtuple('Way', ('id', 'tags', 'refs'))
Relation = namedtuple('Relation', ('id', 'tags', 'members'))


def collectNodes1(filename):
    class NodeHandler(osmium.SimpleHandler):
        def __init__(self):
            osmium.SimpleHandler.__init__(self)
            self.nodes = []

        def node(self, n):
            node_id = n.id
            if node_id % 3 == 1:
                lonlat = (n.location.lon, n.location.lat)
                tags = {tag.k: tag.v for tag in n.tags}
                node = Node(node_id, tags, lonlat)
                self.nodes.append(node)

    node_handler = NodeHandler()
    node_handler.apply_file(filename)
    return 'n', node_handler.nodes


def collectNodes2(filename):
    class NodeHandler(osmium.SimpleHandler):
        def __init__(self):
            osmium.SimpleHandler.__init__(self)
            self.nodes = []

        def node(self, n):
            node_id = n.id
            if node_id % 3 == 2:
                lonlat = (n.location.lon, n.location.lat)
                tags = {tag.k: tag.v for tag in n.tags}
                node = Node(node_id, tags, lonlat)
                self.nodes.append(node)

    node_handler = NodeHandler()
    node_handler.apply_file(filename)
    return 'n', node_handler.nodes


def collectNodes3(filename):
    class NodeHandler(osmium.SimpleHandler):
        def __init__(self):
            osmium.SimpleHandler.__init__(self)
            self.nodes = []

        def node(self, n):
            node_id = n.id
            if node_id % 3 == 0:
                lonlat = (n.location.lon, n.location.lat)
                tags = {tag.k: tag.v for tag in n.tags}
                node = Node(node_id, tags, lonlat)
                self.nodes.append(node)

    node_handler = NodeHandler()
    node_handler.apply_file(filename)
    return 'n', node_handler.nodes


def collectWaysRelations(filename):
    class WayRelationHandler(osmium.SimpleHandler):
        def __init__(self):
            osmium.SimpleHandler.__init__(self)
            self.ways = []
            self.relations = []

        def way(self, w):
            way_id = w.id
            refs = [node.ref for node in w.nodes]
            tags = {tag.k: tag.v for tag in w.tags}
            way = Way(way_id, tags, refs)
            self.ways.append(way)

        def relation(self, r):
            relation_id = r.id
            members = [(member.ref, member.type, member.role) for member in r.members]
            tags = {tag.k: tag.v for tag in r.tags}
            relation = Relation(relation_id, tags, members)
            self.relations.append(relation)

    wr_handler = WayRelationHandler()
    wr_handler.apply_file(filename)
    return 'wr', wr_handler.ways, wr_handler.relations


def readOSMFile(filename):

    bounds = default_bounds

    f = osmium.io.Reader(filename)
    header = f.header()
    print('Bbox:', header.box())

    results = []
    p = mp.Pool(4)
    results.append(p.apply_async(collectNodes1, (filename,)))
    results.append(p.apply_async(collectNodes2, (filename,)))
    results.append(p.apply_async(collectNodes3, (filename,)))
    results.append(p.apply_async(collectWaysRelations, (filename,)))
    p.close()
    p.join()

    nodes, ways, relations = [], [], []
    for result in results:
        result_tuple = result.get()
        if result_tuple[0] == 'n':
            nodes += result_tuple[1]
        elif result_tuple[0] == 'wr':
            ways = result_tuple[1]
            relations = result_tuple[2]

    return {'bounds':bounds, 'nodes':nodes, 'ways':ways, 'relations':relations}



def readCSVFile(folder, encoding):
    node_filepath = os.path.join(folder,'node.csv')
    link_filepath = os.path.join(folder,'link.csv')

    if encoding is None:
        nfin = open(node_filepath, 'r')
        lfin = open(link_filepath, 'r')
    else:
        nfin = open(node_filepath, 'r', encoding=encoding)
        lfin = open(link_filepath, 'r', encoding=encoding)

    reader = csv.DictReader(nfin)
    node_data = [line for line in reader]
    nfin.close()

    reader = csv.DictReader(lfin)
    link_data = [line for line in reader]
    lfin.close()

    return node_data, link_data




# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 15:34
# @desc         [script description]

from osm2gmns.networkclass.basenet import BaseNode, BaseLink, BaseNetwork


class MesoNode(BaseNode):
    def __init__(self, node_id):
        super().__init__(node_id)
        self.macronode = None
        self.macrolink = None
    @property
    def zone_id(self):
        return self.macronode.zone_id if self.macronode is not None else None
    @property
    def macro_node_id(self):
        return self.macronode.node_id if self.macronode is not None else None
    @property
    def macro_link_id(self):
        return self.macrolink.link_id if self.macrolink is not None else None
    @property
    def activity_type(self):
        return self.macronode.activity_type if self.macronode is not None else None
    @property
    def is_boundary(self):
        if self.macronode is None:
            if self.macrolink.from_node.is_boundary is None:
                return None
            else:
                return 0
        else:
            if self.macronode.is_boundary is None:
                return None
            elif self.macronode.is_boundary != 2:
                return self.macronode.is_boundary
            else:
                return -1 if self.incoming_link_list else 1


class MesoLink(BaseLink):
    def __init__(self, link_id):
        super().__init__(link_id)
        self.macronode = None
        self.macrolink = None
        self.movement = None

        self.lanes = 0
        self.lanes_change = []

        self.micronode_list = []  # micronode, lane by lane;
        self.microlink_list = []  # microlink

        self.micronode_bike = []
        self.microlink_bike = []
        self.micronode_walk = []
        self.microlink_walk = []

    @property
    def upstream_normal_link_of_movement_link(self):
        return self.from_node.incoming_link_list[0]
    @property
    def macro_node_id(self):
        return self.macronode.node_id if self.macronode is not None else None
    @property
    def macro_link_id(self):
        return self.macrolink.link_id if self.macrolink is not None else None
    @property
    def mvmt_txt_id(self):
        return self.movement.mvmt_txt_id if self.movement is not None else None
    @property
    def link_type_name(self):
        if self.macrolink is not None:
            return self.macrolink.link_type_name
        else:
            return self.upstream_normal_link_of_movement_link.link_type_name
    @property
    def link_type(self):
        if self.macrolink is not None:
            return self.macrolink.link_type
        else:
            return self.upstream_normal_link_of_movement_link.link_type
    @property
    def free_speed(self):
        if self.macrolink is not None:
            return self.macrolink.free_speed
        else:
            return self.upstream_normal_link_of_movement_link.free_speed
    @property
    def capacity(self):
        if self.macrolink is not None:
            return self.macrolink.capacity
        else:
            return self.upstream_normal_link_of_movement_link.capacity
    @property
    def allowed_uses(self):
        if self.macrolink is not None:
            return self.macrolink.allowed_uses
        else:
            return self.upstream_normal_link_of_movement_link.allowed_uses



class MesoNetwork(BaseNetwork):
    def __init__(self):
        super().__init__()


from osm2gmns.networkclass.macronet import Node, Link, POI
from osm2gmns.osmnet.osmclasses import Way, Relation
from osm2gmns.utils.util import getLogger
from osm2gmns.utils.util_geo import getLineFromNodes, getPolygonFromNodes
import osm2gmns.settings as og_settings
from shapely import geometry
import numpy as np
import random



def _parseRelations(relations, network):
    POI_relation_list = []
    for osm_relation in relations:
        relation = Relation()
        relation.osm_relation_id = str(osm_relation.id)

        for member_id, member_type, member_role in osm_relation.members:
            member_id_str = str(member_id)
            member_type_lc = member_type.lower()
            if member_type_lc == 'node':
                try:
                    relation.member_list.append(network.osm_node_dict[member_id_str])
                except KeyError:
                    logger = getLogger()
                    if logger: logger.info(f'ref node {member_id} in relation {relation.osm_relation_id} is not defined')
            elif member_type_lc == 'way':
                try:
                    relation.member_list.append(network.osm_way_dict[member_id_str])
                except KeyError:
                    logger = getLogger()
                    if logger: logger.info(f'ref way {member_id} in relation {relation.osm_relation_id} is not defined')
            elif member_type_lc == 'relation':
                pass
            else:
                logger = getLogger()
                if logger: logger.warning(f'new member type at relation {relation.osm_relation_id}, {member_type}')

            relation.member_role_list.append(member_role)

        tags = osm_relation.tags
        if 'building' in tags.keys():
            relation.building = tags['building']
        if 'amenity' in tags.keys():
            relation.amenity = tags['amenity']
        if 'name' in tags.keys():
            relation.name = tags['name']

        if relation.building or relation.amenity:
            POI_relation_list.append(relation)
    return POI_relation_list


def _POIFromWay(POI_way_list, net_bound):
    POI_list1 = []
    for way in POI_way_list:
        poly, poly_xy = getPolygonFromNodes(way.ref_node_list)
        if poly is None: continue
        if poly.disjoint(net_bound): continue

        poi = POI()
        poi.osm_way_id = way.osm_way_id
        poi.name = way.name
        poi.building = way.building
        poi.amenity = way.amenity
        poi.leisure = way.leisure
        poi.way = way.way_poi

        poi.geometry, poi.geometry_xy = poly, poly_xy
        lon, lat = poi.geometry.centroid.x, poi.geometry.centroid.y
        poi.centroid = geometry.Point((round(lon,og_settings.lonlat_coord_precision),round(lat,og_settings.lonlat_coord_precision)))
        x, y = poi.geometry_xy.centroid.x, poi.geometry_xy.centroid.y
        poi.centroid_xy = geometry.Point((round(x,og_settings.local_coord_precision),round(y,og_settings.local_coord_precision)))
        POI_list1.append(poi)
    return POI_list1


def _POIFromRelation(POI_relation_list, net_bound):
    POI_list2 = []
    for relation in POI_relation_list:
        poi = POI()
        poi.osm_relation_id = relation.osm_relation_id
        poi.name = relation.name
        poi.building = relation.building
        poi.amenity = relation.amenity
        poi.leisure = relation.leisure
        polygon_list = []
        polygon_list_xy = []
        number_of_members = len(relation.member_list)
        m_ref_node_list = []
        for m in range(number_of_members):
            member = relation.member_list[m]
            role = relation.member_role_list[m]
            if isinstance(member, Way):
                if role != 'outer': continue
                if m_ref_node_list:
                    combined_ref_node_list = []
                    if m_ref_node_list[-1] is member.ref_node_list[0]:
                        combined_ref_node_list = m_ref_node_list + member.ref_node_list[1:]
                    elif m_ref_node_list[-1] is member.ref_node_list[-1]:
                        combined_ref_node_list = m_ref_node_list + list(reversed(member.ref_node_list[:-1]))
                    elif m_ref_node_list[0] is member.ref_node_list[0]:
                        combined_ref_node_list = list(reversed(m_ref_node_list)) + member.ref_node_list[1:]
                    elif m_ref_node_list[0] is member.ref_node_list[-1]:
                        combined_ref_node_list = list(reversed(m_ref_node_list)) + list(reversed(member.ref_node_list[:-1]))

                    if combined_ref_node_list:
                        if combined_ref_node_list[0] is combined_ref_node_list[-1]:
                            poly, poly_xy = getPolygonFromNodes(combined_ref_node_list)
                            if poly is not None:
                                polygon_list.append(poly)
                                polygon_list_xy.append(poly_xy)
                            m_ref_node_list = []
                        else:
                            m_ref_node_list = combined_ref_node_list
                    else:
                        poly, poly_xy = getPolygonFromNodes(m_ref_node_list)
                        if poly is not None:
                            polygon_list.append(poly)
                            polygon_list_xy.append(poly_xy)
                        if member.ref_node_list[0] is member.ref_node_list[-1]:
                            poly, poly_xy = getPolygonFromNodes(m_ref_node_list)
                            if poly is not None:
                                polygon_list.append(poly)
                                polygon_list_xy.append(poly_xy)
                            m_ref_node_list = []
                        else:
                            m_ref_node_list = member.ref_node_list
                else:
                    if member.ref_node_list[0] is member.ref_node_list[-1]:
                        poly, poly_xy = getPolygonFromNodes(member.ref_node_list)
                        if poly is not None:
                            polygon_list.append(poly)
                            polygon_list_xy.append(poly_xy)
                    else:
                        m_ref_node_list = member.ref_node_list
            else:
                logger = getLogger()
                if logger: logger.warning(f'node member detected at building {relation.osm_relation_id}')

        if m_ref_node_list:
            poly, poly_xy = getPolygonFromNodes(m_ref_node_list)
            if poly is not None:
                polygon_list.append(poly)
                polygon_list_xy.append(poly_xy)

        if len(polygon_list) == 0:
            continue
        else:
            if len(polygon_list) == 1:
                poi.geometry = polygon_list[0]
                if poi.geometry.disjoint(net_bound): continue
                poi.geometry_xy = polygon_list_xy[0]
            else:
                disjoint = True
                for poly in polygon_list:
                    if not poly.disjoint(net_bound):
                        disjoint = False
                        break
                if disjoint: continue
                poi.geometry = geometry.MultiPolygon(polygon_list)
                poi.geometry_xy = geometry.MultiPolygon(polygon_list_xy)

            lon, lat = poi.geometry.centroid.x, poi.geometry.centroid.y
            poi.centroid = geometry.Point((round(lon, og_settings.lonlat_coord_precision), round(lat, og_settings.lonlat_coord_precision)))
            x, y = poi.geometry_xy.centroid.x, poi.geometry_xy.centroid.y
            poi.centroid_xy = geometry.Point((round(x, og_settings.local_coord_precision), round(y, og_settings.local_coord_precision)))

            POI_list2.append(poi)

    return POI_list2


def generatePOIs(POI_way_list, osm_relation_list, network, POI_percentage):
    if og_settings.verbose:
        print('    generating POIs')

    POI_list1 = _POIFromWay(POI_way_list, network.bounds)
    POI_list2 = _POIFromRelation(osm_relation_list, network.bounds)

    POI_list_ = POI_list1 + POI_list2
    if POI_percentage == 1:
        POI_list = POI_list_
    else:
        POI_list = random.sample(POI_list_,int(len(POI_list_)*POI_percentage))

    max_poi_id = network.max_poi_id
    for poi in POI_list:
        poi.poi_id = max_poi_id
        max_poi_id += 1
    network.max_poi_id = max_poi_id
    network.POI_list = POI_list


def _findNearestNode(network):
    coord_list = []
    idx_to_node_dict = {}
    for idx, (node_id, node) in enumerate(network.node_dict.items()):
        coord_list.append((node.geometry_xy.x, node.geometry_xy.y))
        idx_to_node_dict[idx] = node
    coord_array = np.array(coord_list)

    for poi in network.POI_list:
        coord = np.array((poi.centroid_xy.x,poi.centroid_xy.y))
        coord_diff = coord_array - coord
        coord_diff_square = np.power(coord_diff,2)
        coord_diff_sum_square = coord_diff_square.sum(axis=1)
        distance = np.sqrt(coord_diff_sum_square)
        idx = int(np.argmin(distance))
        poi.nearest_node = idx_to_node_dict[idx]


def _createConnector(from_node, to_node, link_id):
    link = Link(link_id)
    link.link_type_name = 'connector'
    link.link_type = og_settings.link_type_no_dict[link.link_type_name]
    link.free_speed = og_settings.default_speed_dict[link.link_type_name]
    link.allowed_uses = ['auto','bike','walk']
    link.from_bidirectional_way = True
    link.lanes_list = [og_settings.default_lanes_dict[link.link_type_name]]
    # link.lanes = link.lanes_list[0]
    link.from_node = from_node
    link.to_node = to_node
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)
    link.geometry, link.geometry_xy = getLineFromNodes([link.from_node, link.to_node])
    return link


def _addPOIs(network):
    for poi in network.POI_list:
        node = Node(network.max_node_id)
        node.geometry = poi.geometry.centroid
        node.geometry_xy = poi.geometry_xy.centroid
        node.is_crossing = True
        node.poi_id = poi.poi_id
        network.node_dict[node.node_id] = node
        network.max_node_id += 1

        link1 = _createConnector(node,poi.nearest_node,network.max_link_id)
        network.link_dict[link1.link_id] = link1
        network.max_link_id += 1
        link2 = _createConnector(poi.nearest_node, node, network.max_link_id)
        network.link_dict[link2.link_id] = link2
        network.max_link_id += 1


def connectPOIWithNet(network):
    """
    Connect POIs with the traffic network. Specifically, for each POI, osm2gmns will build a bi-directional connector to connect the POI
    with its nearest node in the traffic network

    Parameters
    ----------
    network: Network
        an osm2gmns Network object

    Returns
    -------
    None
    """
    if len(network.POI_list) == 0:
        print('No POIs found in the network. Please set POIs=True when creating network from osm file')
        return

    _findNearestNode(network)
    _addPOIs(network)










from osm2gmns.movement.connection_from_turns import *
from osm2gmns.movement.auto_connection import *




def generateMovements(net, auto_connection=True):
    processed_node_set = generateMovementsFromTurns(net)

    if auto_connection:
        guessMovements(net, processed_node_set)


    pass


import logging


log = False
log_level = logging.WARNING
log_name = 'osm2gmns'

verbose = True

lonlat_coord_precision = 7
local_coord_precision = 2


default_int_buffer = 20.0           # meter

osm_highway_type_dict = {'motorway': ('motorway', False),
                         'motorway_link': ('motorway', True),
                         'trunk': ('trunk', False),
                         'trunk_link': ('trunk', True),
                         'primary': ('primary', False),
                         'primary_link': ('primary', True),
                         'secondary': ('secondary', False),
                         'secondary_link': ('secondary', True),
                         'tertiary': ('tertiary', False),
                         'tertiary_link': ('tertiary', True),
                         'residential': ('residential', False),
                         'residential_link': ('residential', True),
                         'living_street': ('living_street', False),
                         'service': ('service', False),
                         'services': ('service', False),
                         'cycleway': ('cycleway', False),
                         'footway': ('footway', False),
                         'pedestrian': ('footway', False),
                         'steps': ('footway', False),
                         'track': ('track', False),
                         'unclassified': ('unclassified', False)}

link_type_no_dict = {'motorway':1, 'trunk':2, 'primary':3, 'secondary':4, 'tertiary':5, 'residential':6, 'living_street':7, 'service':8,
                     'cycleway':9, 'footway':10, 'track':11, 'unclassified':15, 'connector':20, 'railway':30, 'aeroway':31}

default_lanes_dict = {'motorway': 4, 'trunk': 3, 'primary': 3, 'secondary': 2, 'tertiary': 2, 'residential': 1, 'living_street':1, 'service': 1,
                      'cycleway':1, 'footway':1, 'track':1, 'unclassified': 1, 'connector': 2}
default_speed_dict = {'motorway': 120, 'trunk': 100, 'primary': 80, 'secondary': 60, 'tertiary': 40, 'residential': 30, 'living_street':30, 'service': 30,
                      'cycleway':5, 'footway':5, 'track':30, 'unclassified': 30, 'connector':120}
default_capacity_dict = {'motorway': 2300, 'trunk': 2200, 'primary': 1800, 'secondary': 1600, 'tertiary': 1200, 'residential': 1000, 'living_street':1000, 'service': 800,
                         'cycleway':800, 'footway':800, 'track':800, 'unclassified': 800, 'connector':9999}


default_oneway_flag_dict = {'motorway': False,'trunk':False,'primary':False,'secondary':False,'tertiary':False,
                            'residential':False,'living_street':False,'service':False,'cycleway':True, 'footway':True,'track': True,
                            'unclassified':False, 'connector':False, 'railway':True, 'aeroway':True}


default_bounds = {'minlat':-90.0, 'minlon':-180.0, 'maxlat':90.0, 'maxlon':180.0}




def _draw(network, figsize):
    import matplotlib.pyplot as plt
    import numpy as np

    xy_list = []
    for node_id, node in network.node_dict.items():
        xy = list(node.geometry_xy.coords)[0]
        xy_list.append(xy)
    xy_array = np.array(xy_list)

    if figsize is None:
        net_length = xy_array[:,0].max() - xy_array[:,0].min()
        net_width = xy_array[:,1].max() - xy_array[:,1].min()
        fig_length = 16
        fig_width = fig_length / net_length * net_width
        plt.figure(figsize=(fig_length, fig_width))
    else:
        plt.figure(figsize=figsize)

    ax = plt.gca()
    ax.axes.xaxis.set_visible(False)
    ax.axes.yaxis.set_visible(False)

    plt.scatter(xy_array[:, 0], xy_array[:, 1], s=16.0, color='darkorange', zorder=1)

    for link_id, link in network.link_dict.items():
        xys = list(link.geometry_xy.coords)
        xys_array = np.array(xys)
        plt.plot(xys_array[:,0], xys_array[:,1], linewidth=1, color='deepskyblue', zorder=0)


def show(network, save=False, figsize=None):
    """
    Show the network in a pop-up window

    Parameters
    ----------
    network: Network
        an osm2gmns Network object
    save: bool
        if True, the plot will also be saved to a local file named network.jpg
    figsize: tuple of int/float, list of int/float
        size of the figure

    Returns
    -------
    None
    """
    if network.number_of_nodes == 0:
        print('WARNING: Cannot show an empty network')
        return

    try:
        import matplotlib.pyplot as plt
    except ImportError as e:
        raise ImportError("Matplotlib is required to show the network") from e

    _draw(network, figsize)
    if save:
        plt.savefig('network.jpg')
        print(f'Figure is saved to network.jpg')

    plt.show()
    plt.close()


def saveFig(network, picpath='network.jpg',figsize=None):
    """
    Save the network plot to a local file

    Parameters
    ----------
    network: Network
        an osm2gmns Network object
    picpath: str
        path to store to network plot. can be an absolute or a relative path
    figsize: tuple of int/float, list of int/float
        size of the figure

    Returns
    -------
    None
    """

    if network.number_of_nodes == 0:
        print('WARNING: Cannot show an empty network')
        return

    try:
        import matplotlib.pyplot as plt
    except ImportError as e:
        raise ImportError("Matplotlib is required to show the network") from e

    _draw(network, figsize)
    plt.savefig(picpath)
    plt.close()
    print(f'Figure is saved to {picpath}')



import osm2gmns.settings as og_settings
import logging


_log_created = False


def getLogger():
    global _log_created

    if not og_settings.log:
        return None

    logger = logging.getLogger(og_settings.log_name)

    if not _log_created:
        logger.setLevel(og_settings.log_level)
        handler = logging.FileHandler('osm2gmns.log', 'w')
        formatter = logging.Formatter('%(asctime)s %(filename)s[line:%(lineno)d] %(levelname)s %(message)s', '%H:%M:%S')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        _log_created = True

    return logger



def config(verbose=og_settings.verbose,
           lonlat_coord_precision=og_settings.lonlat_coord_precision,
           local_coord_precision=og_settings.local_coord_precision,
           log_level=None):
    og_settings.verbose = verbose
    og_settings.lonlat_coord_precision = lonlat_coord_precision
    og_settings.local_coord_precision = local_coord_precision

    if log_level is not None:
        if not isinstance(log_level, str):
            print('WARNING: log_level must be a string and chosen from DEBUG, INFO, WARNING, ERROR, CRITICAL')
        else:
            log_level_upper = log_level.upper()
            if log_level_upper == 'DEBUG':
                og_settings.log = True
                og_settings.log_level = logging.DEBUG
            elif log_level_upper == 'INFO':
                og_settings.log = True
                og_settings.log_level = logging.INFO
            elif log_level_upper == 'WARNING':
                og_settings.log = True
                og_settings.log_level = logging.WARNING
            elif log_level_upper == 'ERROR':
                og_settings.log = True
                og_settings.log_level = logging.ERROR
            elif log_level_upper == 'CRITICAL':
                og_settings.log = True
                og_settings.log_level = logging.CRITICAL
            else:
                print('WARNING: log_level must be a string and chosen from DEBUG, INFO, WARNING, ERROR, CRITICAL')




from osm2gmns.movement.util_mvmt import getLinkAngle
from enum import Enum

Direction = Enum('Direction', ('RIGHTMOST', 'LEFTMOST', 'FORWARD'))

class MainDirections:
    def __init__(self, parent, outgoing, available_lanes):
        pass

    def getStraightest(self):
        return 0

    def empty(self):
        pass

    def includes(self, d):
        pass



def _prepareLinkPriorities(current_link, outgoing, available_lanes):
    priorities = []
    main_directions = MainDirections(current_link, outgoing, available_lanes)
    dist = main_directions.getStraightest()
    if dist == -1:
        return priorities





    return 0


def _divideOnEdges(current_link, connected_links):
    available_lanes = list(range(current_link.lanes))
    priorities = _prepareLinkPriorities(current_link, connected_links, available_lanes)




def _getConnectedSorted(link):
    connected_link_list = []
    for outgoing_link in link.to_node.outgoing_link_list:
        # if outgoing_link in link.forbidden_outgoing_link_list:
        #     continue
        if outgoing_link.to_node is link.from_node:
            continue
        connected_link_list.append(outgoing_link)

    # no need to sort if empty or only with one element
    if len(connected_link_list) <= 1:
        return connected_link_list

    angle_dict = {}
    for connected_link in connected_link_list:
        angle_dict[connected_link] = getLinkAngle(link,connected_link)

    # direction: clockwise
    connected_link_list_sorted = sorted(connected_link_list, key=lambda x: angle_dict[x], reverse=True)
    return connected_link_list_sorted


def _computeLanes2Edges_(current_link):
    connected_links = _getConnectedSorted(current_link)
    if connected_links:
        _divideOnEdges(current_link, connected_links)


def _computeLanes2Edges(link_dict, processed_node_set):
    for link_id, link in link_dict.items():
        if link.to_node.node_id in processed_node_set:
            continue
        _computeLanes2Edges_(link)


def _computeLanes2Lanes(net):
    pass


def guessMovements(net, processed_node_set):
    _computeLanes2Edges(net.link_dict, processed_node_set)
    _computeLanes2Lanes(net)


from osm2gmns.networkclass.macronet import Node, Link, Network
from osm2gmns.io.read_from_osm import readOSMFile
from osm2gmns.osmnet.osmclasses import OSMNode
from osm2gmns.osmnet.combine_links import combineShortLinks
from osm2gmns.osmnet.pois import generatePOIs
from osm2gmns.osmnet.wayfilters import *
from osm2gmns.osmnet.check_args import checkArgs_getNetFromFile
from osm2gmns.utils.util import getLogger
import osm2gmns.settings as og_settings
from shapely import geometry

# todo: remove useless nodes link_type

def _createNodeOnBoundary(node_in, node_outside, network):
    line = network.bounds.intersection(geometry.LineString([node_in.geometry,node_outside.geometry]))
    lon, lat = line.coords[1]
    geometry_lonlat = geometry.Point((round(lon,og_settings.lonlat_coord_precision),round(lat,og_settings.lonlat_coord_precision)))
    boundary_osm_node = OSMNode('', '', geometry_lonlat, True, '', '')
    boundary_osm_node.geometry_xy = network.GT.geo_from_latlon(geometry_lonlat)
    boundary_osm_node.is_crossing = True
    boundary_osm_node.notes = 'boundary node created by osm2gmns'
    return boundary_osm_node


def _getSegmentNodeList(way, segment_no, network):
    m_segment_node_list = way.segment_node_list[segment_no]
    if way.is_reversed: m_segment_node_list = list(reversed(m_segment_node_list))
    number_of_nodes = len(m_segment_node_list)

    m_segment_node_list_group = []

    if m_segment_node_list[0].in_region:
        idx_first_outside = -1
        for idx, node in enumerate(m_segment_node_list):
            if not node.in_region:
                idx_first_outside = idx
                break

        if idx_first_outside == -1:
            m_segment_node_list_group.append(m_segment_node_list)
            return m_segment_node_list_group
        else:
            new_node = _createNodeOnBoundary(m_segment_node_list[idx_first_outside-1],m_segment_node_list[idx_first_outside], network)
            m_segment_node_list_group.append(m_segment_node_list[:idx_first_outside] + [new_node])

    if m_segment_node_list[-1].in_region:
        idx_last_outside = -1
        for idx in range(number_of_nodes-2,-1,-1):
            if not m_segment_node_list[idx].in_region:
                idx_last_outside = idx
                break
        new_node = _createNodeOnBoundary(m_segment_node_list[idx_last_outside+1],m_segment_node_list[idx_last_outside], network)
        m_segment_node_list_group.append([new_node] + m_segment_node_list[idx_last_outside+1:])

    return m_segment_node_list_group


def _createNodeFromOSMNode(network, osmnode):
    if osmnode.node is None:
        node = Node(network.max_node_id)
        node.buildFromOSMNode(osmnode)
        osmnode.node = node
        network.node_dict[node.node_id] = node
        network.max_node_id += 1


def _createNodesAndLinks(network, link_way_list):
    if og_settings.verbose:
        print('    generating nodes and links')

    link_dict = {}
    max_link_id = network.max_link_id
    for way in link_way_list:
        if way.is_pure_cycle: continue
        way.getNodeListForSegments()
        for segment_no in range(way.number_of_segments):
            m_segment_node_list_group = _getSegmentNodeList(way, segment_no, network)
            for m_segment_node_list in m_segment_node_list_group:
                if len(m_segment_node_list) < 2: continue
                _createNodeFromOSMNode(network, m_segment_node_list[0])
                _createNodeFromOSMNode(network, m_segment_node_list[-1])

                link = Link(max_link_id)
                link.buildFromOSMWay(way, 1, m_segment_node_list, network.default_lanes, network.default_speed, network.default_capacity)
                link_dict[link.link_id] = link
                max_link_id += 1
                if not way.oneway:
                    linkb = Link(max_link_id)
                    linkb.buildFromOSMWay(way, -1, list(reversed(m_segment_node_list)), network.default_lanes, network.default_speed, network.default_capacity)
                    link_dict[linkb.link_id] = linkb
                    max_link_id += 1
    network.link_dict = link_dict
    network.max_link_id = max_link_id


def _identifyPureCycleWays(link_way_list):
    for way in link_way_list:
        if way.is_cycle:
            way.is_pure_cycle = True
            for node in way.ref_node_list[1:-1]:
                if node.is_crossing:
                    way.is_pure_cycle = False
                    break


def _addSignalFromLink(network):
    for link_id, link in network.link_dict.items():
        if link.ctrl_type == 'signal':
            to_node = link.to_node
            if to_node.ctrl_type != 'signal':
                if len(to_node.incoming_link_list) > 1 or len(to_node.outgoing_link_list) > 1:
                    to_node.ctrl_type = 'signal'


def _removeIsolated(network, min_nodes):
    if og_settings.verbose:
        print(f'    removing sub networks with less than {min_nodes} nodes')

    node_list = []
    node_to_idx_dict = {}
    for idx, (node_id,node) in enumerate(network.node_dict.items()):
        node_list.append(node)
        node_to_idx_dict[node] = idx

    number_of_nodes = len(node_list)
    node_group_id_list = [-1] * number_of_nodes

    group_id = 0
    start_idx = 0

    while True:
        unprocessed_node_list = [node_list[start_idx]]
        node_group_id_list[start_idx] = group_id
        while unprocessed_node_list:
            node = unprocessed_node_list.pop()
            for ob_link in node.outgoing_link_list:
                ob_node = ob_link.to_node
                if node_group_id_list[node_to_idx_dict[ob_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ob_node]] = group_id
                    unprocessed_node_list.append(ob_node)

            for ib_link in node.incoming_link_list:
                ib_node = ib_link.from_node
                if node_group_id_list[node_to_idx_dict[ib_node]] == -1:
                    node_group_id_list[node_to_idx_dict[ib_node]] = group_id
                    unprocessed_node_list.append(ib_node)

        unreachable_node_exits = False
        idx = 0
        for idx in range(start_idx+1,number_of_nodes):
            if node_group_id_list[idx] == -1:
                unreachable_node_exits = True
                break

        if unreachable_node_exits:
            start_idx = idx
            group_id += 1
        else:
            break

    group_id_set = set(node_group_id_list)
    group_isolated_dict = {}
    for group_id in group_id_set:
        group_size = node_group_id_list.count(group_id)
        if group_size < min_nodes:
            group_isolated_dict[group_id] = True
        else:
            group_isolated_dict[group_id] = False

    removal_link_set = set()
    for idx, node in enumerate(node_list):
        if group_isolated_dict[node_group_id_list[idx]]:
            del network.node_dict[node.node_id]
            for ob_link in node.outgoing_link_list: removal_link_set.add(ob_link)
            for ib_link in node.incoming_link_list: removal_link_set.add(ib_link)
    for link in removal_link_set:
        del network.link_dict[link.link_id]


def _updateDefaultLSC(default_lanes, default_speed, default_capacity, network):
    if default_lanes:
        if isinstance(default_lanes,bool):
            network.default_lanes = og_settings.default_lanes_dict
        elif isinstance(default_lanes,dict):
            network.default_lanes = default_lanes
            for link_type_name, lanes in og_settings.default_lanes_dict.items():
                if link_type_name not in default_lanes.keys():
                    network.default_lanes[link_type_name] = lanes

    if default_speed:
        if isinstance(default_speed,bool):
            network.default_speed = og_settings.default_speed_dict
        elif isinstance(default_speed,dict):
            network.default_speed = default_speed
            for link_type_name, speed in og_settings.default_speed_dict.items():
                if link_type_name not in default_speed.keys():
                    network.default_speed[link_type_name] = speed

    if default_capacity:
        if isinstance(default_capacity,bool):
            network.default_capacity = og_settings.default_capacity_dict
        elif isinstance(default_capacity,dict):
            network.default_capacity = default_capacity
            for link_type_name, capacity in og_settings.default_capacity_dict.items():
                if link_type_name not in default_speed.keys():
                    network.default_capacity[link_type_name] = capacity


def _offsetLinks(network, offset):
    if og_settings.verbose:
        print('    offseting link geometries')

    GT = network.GT

    for link_id, link in network.link_dict.items():
        if link.from_bidirectional_way:
            # distance = max(link.lanes_list) / 2 * 3.5
            geometry_xy = link.geometry_xy.parallel_offset(distance=2, side=offset, join_style=2)
            if offset == 'right':
                link.geometry_xy = geometry.LineString(list(geometry_xy.coords)[::-1])
            elif offset == 'left':
                link.geometry_xy = geometry_xy

            link.geometry = GT.geo_to_latlon(link.geometry_xy)


def _preprocessWays(osmnetwork, link_types, network_types):
    link_way_list = []
    POI_way_list = []
    network_types_set = set(network_types)
    include_railway = True if 'railway' in network_types_set else False
    include_aeroway = True if 'aeroway' in network_types_set else False

    for osm_way_id, way in osmnetwork.osm_way_dict.items():
        if way.building or way.amenity or way.leisure:
            POI_way_list.append(way)

        elif way.highway:
            if way.highway in highway_poi_set:
                way.way_poi = way.highway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.highway in negligible_highway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue

            try:
                way.link_type_name, way.is_link = og_settings.osm_highway_type_dict[way.highway]
                way.link_type = og_settings.link_type_no_dict[way.link_type_name]
            except KeyError:
                logger = getLogger()
                if logger: logger.warning(f'new highway type at way {way.osm_way_id}, {way.highway}')
                continue

            valid_link_type = link_types == 'all' or way.link_type_name in link_types
            if not valid_link_type:
                continue

            allowable_agent_type_list = getAllowableAgentType(way)
            way.allowable_agent_type_list = list(set(allowable_agent_type_list).intersection(network_types_set))
            if len(way.allowable_agent_type_list) == 0:
                continue
            way.allowed_uses = way.allowable_agent_type_list

            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list:
                node.usage_count += 1

            if way.ref_node_list[0] is way.ref_node_list[-1]:
                way.is_cycle = True
            if way.oneway is None:
                if way.junction in ['circular', 'roundabout']:
                    way.oneway = True
                else:
                    way.oneway = og_settings.default_oneway_flag_dict[way.link_type_name]
            way.link_class = 'highway'
            link_way_list.append(way)

        elif way.railway:
            if not include_railway: continue

            if way.railway in railway_poi_set:
                way.way_poi = way.railway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.railway in negligible_railway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue

            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list[1:-1]: node.usage_count += 1

            way.link_type_name = way.railway
            way.link_type = og_settings.link_type_no_dict['railway']
            if way.oneway is None:
                way.oneway = og_settings.default_oneway_flag_dict['railway']
            way.link_class = 'railway'
            link_way_list.append(way)

        elif way.aeroway:
            if not include_aeroway: continue

            if way.aeroway in aeroway_poi_set:
                way.way_poi = way.aeroway
                POI_way_list.append(way)
                continue
            if way.area and way.area != 'no':
                continue
            if way.aeroway in negligible_aeroway_type_set:
                continue
            if len(way.ref_node_list) < 2:
                continue

            way.ref_node_list[0].is_crossing = True
            way.ref_node_list[-1].is_crossing = True
            for node in way.ref_node_list[1:-1]: node.usage_count += 1

            way.link_type_name = way.aeroway
            way.link_type = og_settings.link_type_no_dict['aeroway']
            if way.oneway is None:
                way.oneway = og_settings.default_oneway_flag_dict['aeroway']
            way.link_class = 'aeroway'
            link_way_list.append(way)

        else:
            pass

    osmnetwork.link_way_list = link_way_list
    osmnetwork.POI_way_list = POI_way_list


def _identifyCrossingOSMNodes(osm_node_dict):
    for osm_node_id, osmnode in osm_node_dict.items():
        if osmnode.usage_count >= 2 or osmnode.ctrl_type == 'signal':
            osmnode.is_crossing = True


def _createNLPs(osmnetwork, network, link_types, offset, network_types, POI, POI_percentage):
    _preprocessWays(osmnetwork, link_types, network_types)
    _identifyCrossingOSMNodes(osmnetwork.osm_node_dict)
    _identifyPureCycleWays(osmnetwork.link_way_list)
    _createNodesAndLinks(network, osmnetwork.link_way_list)
    _addSignalFromLink(network)

    if offset != 'no': _offsetLinks(network, offset)

    if POI: generatePOIs(osmnetwork.POI_way_list, osmnetwork.osm_relation_list, network, POI_percentage)



def _buildNet(osmnetwork, network_types, link_types, POI, POI_percentage, offset, min_nodes, combine, default_lanes,
              default_speed, default_capacity, start_node_id, start_link_id):
    if og_settings.verbose:
        print('  parsing osm network')

    network = Network()
    network.max_node_id = start_node_id
    network.max_link_id = start_link_id

    network.GT, network.bounds = osmnetwork.GT, osmnetwork.bounds

    # update default lanes, speed, and capacity
    _updateDefaultLSC(default_lanes, default_speed, default_capacity, network)

    # create node, link, poi
    _createNLPs(osmnetwork, network, link_types, offset, network_types, POI, POI_percentage)

    # remove isolated nodes and links
    if min_nodes > 1: _removeIsolated(network, min_nodes)

    # combine adjacent links at two-degree nodes
    if combine: combineShortLinks(network)

    return network


def getNetFromOSMFile(osm_filename='map.osm', network_types=('auto',), link_types='all', POI=False, POI_sampling_ratio=1.0,
                      strict_mode=True, offset='no', min_nodes=1, combine=False, bbox=None,
                      default_lanes=False, default_speed=False, default_capacity=False):
    """
    Deprecated
    """

    print('WARNING: getNetFromOSMFile() and getNetFromPBFFile() are deprecated and will be removed in a future release.\n'
          '         Please use getNetFromFile().')

    network = getNetFromFile(osm_filename, network_types, link_types, POI, POI_sampling_ratio, strict_mode, offset, min_nodes,
                             combine, bbox, default_lanes, default_speed, default_capacity)
    return network


def getNetFromPBFFile(pbf_filename='map.osm.pbf', network_types=('auto',), link_types='all', POI=False, POI_sampling_ratio=1.0,
                      strict_mode=True, offset='no', min_nodes=1, combine=False, bbox=None,
                      default_lanes=False, default_speed=False, default_capacity=False):
    """
    Deprecated
    """

    print('WARNING: getNetFromOSMFile() and getNetFromPBFFile() are deprecated and will be removed in a future release.\n'
          '         Please use getNetFromFile().')
    network = getNetFromFile(pbf_filename, network_types, link_types, POI, POI_sampling_ratio, strict_mode, offset, min_nodes,
                             combine, bbox, default_lanes, default_speed, default_capacity)
    return network


def getNetFromFile(filename='map.osm', network_types=('auto',), link_types='all', POI=False, POI_sampling_ratio=1.0,
                   strict_mode=True, offset='no', min_nodes=1, combine=False, bbox=None,
                   default_lanes=False, default_speed=False, default_capacity=False, start_node_id=0, start_link_id=0):
    """
    Get an osm2gmns Network object from an osm file

    Parameters
    ----------
    filename: str
        path of an osm file; can be absolute or relative path; supported osm file formats: .osm, .xml, and .pbf
    network_types: str, tuple of strings, list of strings, or set of strings
        osm2gmns supports five different network types, including auto, bike, walk, railway, and aeroway.
        network_types can be any one or any combinations of the five supported network types
    link_types: str, tuple of strings, list of strings, or set of strings
        supported link types: motorway, trunk, primary, secondary, tertiary, residential, service, cycleway,
        footway, track, unclassified, connector, railway, and aeroway.
    POI: bool
        if extract point of interest information
    POI_sampling_ratio: float
        prcentage of POIs to be extracted if POI is set as True. this value should be a float number between 0.0 and 1.0.
    strict_mode: bool
        if True, network elements (node, link, poi) outside the boundary will be discarded
    offset: str
        offset overlapping links. the value of this argument can be 'left', 'right', or 'no'
    min_nodes: int
        a network return by the function may contain several sub-networks that are disconnected from each other.
        sub-networks with the number of nodes less than min_nodes will be discarded
    combine: bool
        if True, adhacent short links with the same attributes will be combined into a long link. the operation will only
        be performed on short links connected with a two-degree nodes (one incoming link and one outgoing link)
    bbox: tuple of four float/int values, list of four float/int values, None
        specify the boundary of the network to be extracted, consisting of minimum latitude, minimum longtitude, maximum latitude, and maximum longitud.
        if None, osm2gmns will try to find network boundary from the input osm file
    default_lanes: bool, dict
        if True, assign a default value for links without lanes information based on built-in settings. if a dict,
        assign a default value for links without lanes information based on the dict passed by users.
    default_speed: bool, dict
        if True, assign a default value for links without speed information based on built-in settings. if a dict,
        assign a default value for links without speed information based on the dict passed by users.
    default_capacity: bool, dict
        if True, assign a default value for links without capacity information based on built-in settings. if a dict,
        assign a default value for links without capacity information based on the dict passed by users.
    start_node_id: int
        osm2gmns assigns node_ids to generated nodes starting from start_node_id.
    start_link_id: int
        osm2gmns assigns link_ids to generated links starting from start_link_id

    Returns
    -------
    network: Network
        osm2gmns Network object
    """

    if og_settings.verbose:
        print('arguments used for network parsing:')
        print(f'  filename: {filename}')
        print(f'  network_types: {network_types}')
        print(f'  link_types: {link_types}')
        print(f'  POI: {POI}')
        print(f'  POI_sampling_ratio: {POI_sampling_ratio}')
        print(f'  strict_mode: {strict_mode}')
        print(f'  offset: {offset}')
        print(f'  min_nodes: {min_nodes}')
        print(f'  combine: {combine}')
        print(f'  bbox: {bbox}')
        print(f'  default_lanes: {default_lanes}')
        print(f'  default_speed: {default_speed}')
        print(f'  default_capacity: {default_capacity}')
        print(f'  start_node_id: {start_node_id}')
        print(f'  start_link_id: {start_link_id}\n')

        print('Building Network from OSM file')

    network_types_, link_types_, POI_, POI_sampling_ratio_, strict_mode_, offset_, min_nodes_, combine_, \
        bbox_, default_lanes_, default_speed_, default_capacity_, start_node_id_, start_link_id_ = \
        checkArgs_getNetFromFile(filename, network_types, link_types, POI, POI_sampling_ratio, strict_mode, offset,
                                min_nodes, combine, bbox, default_lanes, default_speed, default_capacity, start_node_id,
                                 start_link_id)

    osmnetwork = readOSMFile(filename, POI_, strict_mode_, bbox_)

    network = _buildNet(osmnetwork, network_types_, link_types_, POI_, POI_sampling_ratio_, offset_, min_nodes_, combine_,
                        default_lanes_, default_speed_, default_capacity_, start_node_id_, start_link_id_)

    if og_settings.verbose:
        print(f'  number of nodes: {len(network.node_dict)}, number of links: {len(network.link_dict)}, number of pois: {len(network.POI_list)}')

    return network





#!/usr/bin/env python
# Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
# Copyright (C) 2009-2020 German Aerospace Center (DLR) and others.
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# https://www.eclipse.org/legal/epl-2.0/
# This Source Code may also be made available under the following Secondary
# Licenses when the conditions for such availability set forth in the Eclipse
# Public License 2.0 are satisfied: GNU General Public License, version 2
# or later which is available at
# https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later

# @file    osmGet.py
# @author  Daniel Krajzewicz
# @author  Jakob Erdmann
# @author  Michael Behrisch
# @date    2009-08-01


import os
import http.client as httplib
import urllib.parse as urlparse
import base64



_url = "www.overpass-api.de/api/interpreter"
# alternatives: overpass.kumi.systems/api/interpreter, sumo.dlr.de/osm/api/interpreter


def _readCompressed(conn, urlpath, query, filename):
    conn.request("POST", "/" + urlpath, """
    <osm-script timeout="240" element-limit="1073741824">
    <union>
       %s
       <recurse type="node-relation" into="rels"/>
       <recurse type="node-way"/>
       <recurse type="way-relation"/>
    </union>
    <union>
       <item/>
       <recurse type="way-node"/>
    </union>
    <print mode="body"/>
    </osm-script>""" % query)
    response = conn.getresponse()
    # print(response.status, response.reason)
    if response.status == 200:
        print('valid reponses got from API server.')
        print('receving data...')
        out = open(filename, "wb")
        out.write(response.read())
        out.close()
        print(f'map data has been writen to {filename}')


def downloadOSMData(area_id, output_filename='map.osm', url=_url):
    file_name, file_extension = os.path.splitext(output_filename)
    if not file_extension:
        print(f'WARNING: no file extension in output_filename {output_filename}, output_filename is changed to {file_name}.osm')
        output_filename = f'{file_name}.osm'
    elif file_extension not in ['.osm', '.xml']:
        print(f'WARNING:  the file extension in output_filename {output_filename} is not supported, output_filename is changed to {file_name}.osm')
        output_filename = f'{file_name}.osm'

    if "http" in url:
        url = urlparse.urlparse(url)
    else:
        url = urlparse.urlparse("https://" + url)
    if os.environ.get("https_proxy") is not None:
        headers = {}
        proxy_url = urlparse.urlparse(os.environ.get("https_proxy"))
        if proxy_url.username and proxy_url.password:
            auth = '%s:%s' % (proxy_url.username, proxy_url.password)
            headers['Proxy-Authorization'] = 'Basic ' + base64.b64encode(auth)
        conn = httplib.HTTPSConnection(proxy_url.hostname, proxy_url.port)
        conn.set_tunnel(url.hostname, 443, headers)
    else:
        if url.scheme == "https":
            conn = httplib.HTTPSConnection(url.hostname, url.port)
        else:
            conn = httplib.HTTPConnection(url.hostname, url.port)

    if area_id < 3600000000:
        area_id += 3600000000
    _readCompressed(conn, url.path, '<area-query ref="%s"/>' % area_id, output_filename)

    conn.close()



from osm2gmns.osmnet.build_net import getNetFromOSMFile, getNetFromPBFFile, getNetFromFile
from osm2gmns.osmnet.complex_intersection import consolidateComplexIntersections
from osm2gmns.osmnet.enrich_net_info import generateNodeActivityInfo, generateLinkVDFInfo
from osm2gmns.osmnet.pois import connectPOIWithNet
from osm2gmns.osmnet.visualization import show, saveFig
from osm2gmns.movement.generate_movements_old import generateMovements
# from osm2gmns.movement.generate_movements import generateMovements
from osm2gmns.multiresolutionnet.build_mrnet import buildMultiResolutionNets
from osm2gmns.io.load_from_csv import getNetFromCSV, loadNetFromCSV
from osm2gmns.io.downloader import downloadOSMData
from osm2gmns.io.writefile import outputNetToCSV
from osm2gmns.utils.util import config
import osm2gmns.settings as og_settings


__version__ = '0.6.8'
print(f'osm2gmns, {__version__}')



# todo: turns
# todo: parse speed


# todo: link_id: int, str
# todo: compatible with networkx
# todo: mrnet bike walk
# todo: movement generation



import osm2gmns.settings as og_settings


def generateNodeActivityInfo(network):
    """
    Generate activity information, including activity_type, is_boundary, zone_id for nodes. activity_type includes
    motorway, primary, secondary, tertiary, residential, etc, and is determined by adjacent links

    Parameters
    ----------
    network: Network
        osm2gmns Network object

    Returns
    -------
    None
    """

    if og_settings.verbose:
        print('Generating Node Activity Information')

    # node activity
    node_adjacent_link_type_count_dict = {}

    for link_id, link in network.link_dict.items():
        if link.link_type_name == 'connector': continue

        from_node = link.from_node
        if from_node not in node_adjacent_link_type_count_dict.keys():
            node_adjacent_link_type_count_dict[from_node] = {}
        adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[from_node]

        if link.link_type_name in adjacent_link_type_count_dict.keys():
            adjacent_link_type_count_dict[link.link_type_name] += 1
        else:
            adjacent_link_type_count_dict[link.link_type_name] = 1

        to_node = link.to_node
        if to_node not in node_adjacent_link_type_count_dict.keys():
            node_adjacent_link_type_count_dict[to_node] = {}
        adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[to_node]

        if link.link_type_name in adjacent_link_type_count_dict.keys():
            adjacent_link_type_count_dict[link.link_type_name] += 1
        else:
            adjacent_link_type_count_dict[link.link_type_name] = 1

    for node_id, node in network.node_dict.items():
        if node.poi_id is not None:
            node.activity_type = 'poi'
        elif node in node_adjacent_link_type_count_dict.keys():
            adjacent_link_type_count_dict = node_adjacent_link_type_count_dict[node]
            max_count_type = ''
            max_count = 0
            for link_type_name, count in adjacent_link_type_count_dict.items():
                if count > max_count:
                    max_count = count
                    max_count_type = link_type_name
            node.activity_type = max_count_type

    # boundary
    for node_id, node in network.node_dict.items():
        node.is_boundary = 0
        if node.activity_type == 'poi': continue

        if len(node.outgoing_link_list) == 0:
            node.is_boundary = -1
        elif len(node.incoming_link_list) == 0:
            node.is_boundary = 1
        elif (len(node.incoming_link_list) == 1) and (len(node.outgoing_link_list) == 1):
            ib_link = node.incoming_link_list[0]
            ob_link = node.outgoing_link_list[0]
            if ib_link.from_node is ob_link.to_node:
                node.is_boundary = 2

    # zone
    for node_id, node in network.node_dict.items():
        if node.is_boundary == 0: continue
        node.zone_id = node.node_id



def generateLinkVDFInfo(network):
    """
    Generate VDF information, including VDF_fftt1 and VDF_cap1 for links.
    The unit of VDF_fftt1 and VDF_cap1 are min and veh/hour/link, respectively

    Parameters
    ----------
    network: Network
        osm2gmns Network object

    Returns
    -------
    None
    """

    if og_settings.verbose:
        print('Generating Link VDF Information')

    for link_id, link in network.link_dict.items():
        if link.capacity is None:
            lanes = og_settings.default_lanes_dict[link.link_type_name] if link.lanes is None else link.lanes
            link.VDF_cap1 = lanes * og_settings.default_capacity_dict[link.link_type_name]
        else:
            link.VDF_cap1 = link.capacity

        if link.free_speed is None:
            free_speed = og_settings.default_speed_dict[link.link_type_name]
        else:
            free_speed = link.free_speed
        link.VDF_FFTT1 = link.length / free_speed * 0.06



# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 15:34
# @desc         [script description]

from osm2gmns.networkclass.basenet import BaseNode, BaseLink, BaseNetwork

class MicroNode(BaseNode):
    def __init__(self, node_id):
        super().__init__(node_id)
        self.mesolink = None
        self.lane_no = 0
        self.is_link_upstream_end_node = False        # nodes on the two ends of macro links
        self.is_link_downstream_end_node = False

    @property
    def zone_id(self):
        if self.is_link_upstream_end_node:
            return self.mesolink.macrolink.from_node.zone_id
        elif self.is_link_downstream_end_node:
            return self.mesolink.macrolink.to_node.zone_id
        else:
            return None

    @property
    def is_boundary(self):
        if self.mesolink.from_node.is_boundary is None:
            return None
        else:
            if self.is_link_upstream_end_node:
                return self.mesolink.from_node.is_boundary
            elif self.is_link_downstream_end_node:
                return self.mesolink.to_node.is_boundary
            else:
                return 0


class MicroLink(BaseLink):
    def __init__(self, link_id):
        super().__init__(link_id)
        self.mesolink = None
        self.cell_type = 1	            # //1:traveling; 2:changing
        self.allowed_uses = []
        self.is_first_movement_cell = False

    @property
    def mvmt_txt_id(self):
        if self.is_first_movement_cell and self.mesolink.mvmt_txt_id is not None:
            return self.mesolink.mvmt_txt_id
    @property
    def lane_no(self):
        return self.from_node.lane_no



class MicroNetwork(BaseNetwork):
    def __init__(self):
        super().__init__()


from osm2gmns.io.output_mrnet import outputMesoNet, outputMicroNet
import osm2gmns.settings as og_settings
from shapely import wkt
import csv
import os


# we use errors='ignore' to make our program compatible with some characters that cannot be encoded by the local encoding,
# these characters will be discarded

def _outputNode(network, output_folder, node_filename, projection, encoding):
    node_filepath = os.path.join(output_folder, node_filename)
    while True:
        try:
            if encoding is None:
                outfile = open(node_filepath, 'w', newline='',errors='ignore')
            else:
                outfile = open(node_filepath, 'w', newline='', errors='ignore', encoding=encoding)
            break
        except PermissionError:
            print(f'{node_filename} may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'node_id', 'osm_node_id', 'osm_highway', 'zone_id', 'ctrl_type', 'node_type', 'activity_type',
                     'is_boundary', 'x_coord', 'y_coord', 'intersection_id','poi_id','notes'])
    for node_id,node in network.node_dict.items():
        if projection:
            x_coord = round(node.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(node.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(node.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(node.geometry.y, og_settings.lonlat_coord_precision)
        line = [node.name, node.node_id, node.osm_node_id, node.osm_highway, node.zone_id, node.ctrl_type, '', node.activity_type,
                node.is_boundary, x_coord, y_coord,node.intersection_id,node.poi_id,node.notes]
        writer.writerow(line)
    outfile.close()


def _outputLink(network, output_folder, link_filename, projection, encoding):
    link_filepath = os.path.join(output_folder, link_filename)
    while True:
        try:
            if encoding is None:
                outfile = open(link_filepath, 'w', newline='',errors='ignore')
            else:
                outfile = open(link_filepath, 'w', newline='', errors='ignore', encoding=encoding)
            break
        except PermissionError:
            print(f'{link_filename} may be locked by other programs. please release it then press Enter to try again')
            input()
    writer = csv.writer(outfile)
    writer.writerow(['name', 'link_id', 'osm_way_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes',
                     'free_speed', 'capacity', 'link_type_name', 'link_type', 'geometry','allowed_uses','from_biway',
                     'is_link', 'VDF_fftt1', 'VDF_cap1'])
    for link_id, link in network.link_dict.items():
        from_biway = 1 if link.from_bidirectional_way else 0
        is_link = 1 if link.is_link else 0
        if projection:
            geometry_ = wkt.dumps(link.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(link.geometry, rounding_precision=og_settings.lonlat_coord_precision)
        line = [link.name, link.link_id, link.osm_way_id, link.from_node.node_id, link.to_node.node_id, link.dir_flag, link.length,
                link.lanes, link.free_speed, link.capacity, link.link_type_name, link.link_type, geometry_, ';'.join(link.allowed_uses),
                from_biway, is_link, link.VDF_fftt1, link.VDF_cap1]
        writer.writerow(line)
    outfile.close()


def _outputMovement(network, output_folder, movement_filename, projection, encoding):
    movement_filepath = os.path.join(output_folder, movement_filename)
    if network.max_movement_id > 0:
        while True:
            try:
                if encoding is None:
                    outfile = open(movement_filepath, 'w', newline='', errors='ignore')  # ,errors='ignore'
                else:
                    outfile = open(movement_filepath, 'w', newline='', errors='ignore',
                                   encoding=encoding)  # ,errors='ignore'
                break
            except PermissionError:
                print(
                    f'{movement_filename} may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['mvmt_id', 'node_id', 'osm_node_id', 'name', 'ib_link_id', 'start_ib_lane', 'end_ib_lane',
                         'ob_link_id', 'start_ob_lane', 'end_ob_lane', 'lanes', 'ib_osm_node_id', 'ob_osm_node_id',
                         'type', 'penalty', 'capacity', 'ctrl_type', 'mvmt_txt_id', 'geometry', 'volume', 'free_speed',
                         'allowed_uses', 'generated_by_osm2gmns'])

        for node_id, node in network.node_dict.items():
            for mvmt in node.movement_list:
                from_node = mvmt.ib_link.from_node
                to_node = mvmt.ob_link.to_node
                if projection:
                    geometry_ = wkt.dumps(mvmt.geometry_xy, rounding_precision=og_settings.local_coord_precision)
                else:
                    geometry_ = wkt.dumps(mvmt.geometry, rounding_precision=og_settings.lonlat_coord_precision)

                generated_by_osm2gmns = 1 if mvmt.generated_by_osm2gmns else 0

                line = [mvmt.movement_id, node.node_id, node.osm_node_id, '', mvmt.ib_link.link_id, mvmt.start_ib_lane,
                        mvmt.end_ib_lane, mvmt.ob_link.link_id, mvmt.start_ob_lane, mvmt.end_ob_lane, mvmt.lanes,
                        from_node.osm_node_id, to_node.osm_node_id, mvmt.type, '', '', mvmt.ctrl_type, mvmt.mvmt_txt_id,
                        geometry_, '', '', ';'.join(mvmt.allowed_uses), generated_by_osm2gmns]
                writer.writerow(line)
        outfile.close()


def _outputPOI(network, output_folder, poi_filename, projection, encoding):
    poi_filepath = os.path.join(output_folder, poi_filename)
    if network.POI_list:
        while True:
            try:
                if encoding is None:
                    outfile = open(poi_filepath, 'w', newline='',errors='ignore')     # ,errors='ignore'
                else:
                    outfile = open(poi_filepath, 'w', newline='', errors='ignore', encoding=encoding)  # ,errors='ignore'
                break
            except PermissionError:
                print(f'{poi_filename} may be locked by other programs. please release it then press Enter to try again')
                input()
        writer = csv.writer(outfile)
        writer.writerow(['name', 'poi_id', 'osm_way_id', 'osm_relation_id','building','amenity','leisure','way','geometry','centroid',
                         'area', 'area_ft2'])
        for poi in network.POI_list:
            geometry_ = poi.geometry_xy if projection else poi.geometry
            centroid = poi.centroid_xy if projection else poi.centroid
            area = poi.geometry_xy.area
            area_ft2 = area * 10.7639
            line = [poi.name, poi.poi_id, poi.osm_way_id, poi.osm_relation_id, poi.building, poi.amenity, poi.leisure, poi.way, geometry_,
                    centroid, round(area,1), round(area_ft2,1)]
            writer.writerow(line)


def outputNetToCSV(network, output_folder='', prefix='', projection=False, encoding=None):
    """
    Output an osm2gmns network object to csv files in GMNS format

    Parameters
    ----------
    network: Network
        an osm2gmns network object
    output_folder: str
        path of the folder to store network files. can be an absolute or a relative path
    prefix: str
        prefix of output csv files
    projection: bool
        if True, osm2gmns will project the network to a local coordinate system when ouptting a network
    encoding: str
        the file encoding used to output a network

    Returns
    -------
    None
    """
    if og_settings.verbose:
        print('Outputting Network Files')

    if output_folder:
        if not os.path.isdir(output_folder): os.mkdir(output_folder)

    node_filename = f'{prefix}node.csv'
    _outputNode(network, output_folder, node_filename, projection, encoding)

    link_filename = f'{prefix}link.csv'
    _outputLink(network, output_folder, link_filename, projection, encoding)

    movement_filename = f'{prefix}movement.csv'
    _outputMovement(network, output_folder, movement_filename, projection, encoding)

    poi_filename = f'{prefix}poi.csv'
    _outputPOI(network, output_folder, poi_filename, projection, encoding)

    if network.mesonet is not None:
        outputMesoNet(network.mesonet, output_folder, prefix, projection, encoding)
    if network.micronet is not None:
        outputMicroNet(network.micronet, output_folder, prefix, projection, encoding)



from osm2gmns.osmnet.wayfilters import network_types_all
import os
import sys


def checkArgs_getNetFromFile(filename, network_types, link_types, POI, POI_sampling_ratio,
               strict_mode, offset, min_nodes, combine, bbox,
               default_lanes, default_speed, default_capacity, start_node_id, start_link_id):
    # filename
    file_extension = os.path.splitext(filename)[-1]
    if not file_extension:
        sys.exit(f'ERROR: cannot detect the format of file {filename}. If it is a xml-based text file, please add extension .xml or .osm after the filename')
    if not os.path.exists(filename):
        sys.exit(f'ERROR: file {filename} does not exist')

    # network_types
    if isinstance(network_types, str):
        network_types_temp = [network_types,]
    elif isinstance(network_types, list) or isinstance(network_types, set) or isinstance(network_types, tuple):
        network_types_temp = network_types
    else:
        sys.exit(f'ERROR: network_types must be a string, list, set, or tuple')

    network_types_ = []
    for net_type in network_types_temp:
        if net_type not in network_types_all:
            print(f'WARNING: network_type \'{net_type}\' does not belong to {network_types_all}, it will be skipped')
        else:
            network_types_.append(net_type)

    # link_types
    if link_types != 'all':
        if isinstance(link_types, str):
            link_types_ = (link_types,)
        elif isinstance(link_types, list) or isinstance(link_types, set) or isinstance(link_types, tuple):
            link_types_ = link_types
        else:
            sys.exit(f'ERROR: argument link_types must be a string, list, set, or tuple')
    else:
        link_types_ = link_types

    # POIs
    if isinstance(POI, bool):
        POI_ = POI
    else:
        print('WARNING: argument POI should be a bool. The default value False has been applied')
        POI_ = False

    # POI_sampling_ratio
    if isinstance(POI_sampling_ratio, int) or isinstance(POI_sampling_ratio, float):
        if POI_sampling_ratio < 0.0 or POI_sampling_ratio > 1.0:
            print('WARNING: argument POI_sampling_ratio should be a float between 0.0 and 1.0. The default value 1.0 has been applied')
            POI_sampling_ratio_ = 1.0
        else:
            POI_sampling_ratio_ = POI_sampling_ratio
    else:
        print('WARNING: argument POI_sampling_ratio should be a float between 0.0 and 1.0. The default value 1.0 has been applied')
        POI_sampling_ratio_ = 1.0

    # strict_mode
    if isinstance(strict_mode, bool):
        strict_mode_ = strict_mode
    else:
        print('WARNING: argument strict_mode should be a bool. The default value True has been applied')
        strict_mode_ = True

    # offset
    if offset in ('left', 'right', 'no'):
        offset_ = offset
    else:
        print("WARNING: argument offset should be chosen from 'left', 'right', 'no'. The default value 'no' has been applied")
        offset_ = 'no'

    # min_nodes
    if isinstance(min_nodes, int):
        if min_nodes < 1:
            print('WARNING: argument min_nodes should be a integer greater than or equal to 1. The default value 1 has been applied')
            min_nodes_ = 1
        else:
            min_nodes_ = min_nodes
    else:
        print('WARNING: argument min_nodes should be an integer greater than or equal to 1. The default value 1 has been applied')
        min_nodes_ = 1

    # combine
    if isinstance(combine, bool):
        combine_ = combine
    else:
        print('WARNING: argument combine should be a bool. The default value False has been applied')
        combine_ = False

    # bbox
    if bbox is None:
        bbox_ = bbox
    elif isinstance(bbox, tuple) or isinstance(bbox, list):
        if len(bbox) == 4:
            for ll in bbox:
                if not(isinstance(ll, int) or isinstance(ll, float)):
                    sys.exit('ERROR: argument bbox should be a tuple with four floats: (minlat, minlon, maxlat, maxlon)')
            bbox_ = bbox
        else:
            sys.exit('ERROR: argument bbox should be a tuple with four floats: (minlat, minlon, maxlat, maxlon)')
    else:
        sys.exit('ERROR: argument bbox should be a tuple with four floats: (minlat, minlon, maxlat, maxlon)')

    # default_lanes
    if isinstance(default_lanes, bool) or isinstance(default_lanes, dict):
        default_lanes_ = default_lanes
    else:
        print('WARNING: argument default_lanes should be a bool or a dict. The default value False has been applied')
        default_lanes_ = False

    # default_speed
    if isinstance(default_speed, bool) or isinstance(default_speed, dict):
        default_speed_ = default_speed
    else:
        print('WARNING: argument default_speed should be a bool or a dict. The default value False has been applied')
        default_speed_ = False

    # default_capacity
    if isinstance(default_capacity, bool) or isinstance(default_capacity, dict):
        default_capacity_ = default_capacity
    else:
        print('WARNING: argument default_capacity should be a bool or a dict. The default value False has been applied')
        default_capacity_ = False

    # start node/link id
    if isinstance(start_node_id, int):
        start_node_id_ = start_node_id
    else:
        print('WARNING: argument start_node_id should be a int. The default value 0 has been applied')
        start_node_id_ = 0

    if isinstance(start_link_id, int):
        start_link_id_ = start_link_id
    else:
        print('WARNING: argument start_link_id should be a int. The default value 0 has been applied')
        start_link_id_ = 0

    return network_types_, link_types_, POI_, POI_sampling_ratio_, strict_mode_, offset_, min_nodes_, combine_,\
        bbox_, default_lanes_, default_speed_, default_capacity_, start_node_id_, start_link_id_


# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/12/8 12:21
# @desc         [script description]


def getFileHandle(filepath, encoding):

    while True:
        try:
            if encoding is None:
                outfile = open(filepath, 'w', newline='',errors='ignore')
            else:
                outfile = open(filepath, 'w', newline='', errors='ignore', encoding=encoding)
            break
        except PermissionError:
            print(f'{filepath} may be locked by other programs. please release it then press Enter to try again')
            input()

    return outfile


from shapely.geometry import Point


class OSMNode:
    def __init__(self, osm_node_name, osm_node_id, geometry, in_region, osm_highway, ctrl_type):
        """
        OSMNode Construction function

        Parameters
        ----------
        osm_node_name: str
        osm_node_id: str
        geometry: Point
        in_region: bool
        osm_highway: str
        ctrl_type: str
        """
        self.name = osm_node_name
        self.osm_node_id = osm_node_id
        self.geometry = geometry
        self.geometry_xy = None
        self.osm_highway = osm_highway
        # self.node_type = ''
        self.ctrl_type = ctrl_type
        self.in_region = in_region
        self.is_crossing = False
        # self.is_isolated = False
        # self.valid = True

        self.notes = ''
        self.node = None

        self.usage_count = 0
        # self.belongs_to_valid_link_type = False


class Way:
    def __init__(self):
        self.osm_way_id = None          # string
        self.highway = None
        self.railway = None
        self.aeroway = None
        self.link_class = ''                    # highway, railway, aeroway
        self.link_type_name = ''
        self.link_type = 0
        self.is_link = False
        self.name = None
        self.lanes = None
        self.forward_lanes = None
        self.backward_lanes = None
        self.turn_lanes = None
        self.turn_lanes_forward = None
        self.turn_lanes_backward = None
        self.maxspeed = None
        self.oneway = None

        self.junction = None
        self.area = None
        self.motor_vehicle = None
        self.motorcar = None
        self.service = None
        self.access = None
        self.foot = None
        self.bicycle = None
        self.building = None
        self.amenity = None
        self.leisure = None
        self.way_poi = None     # used for poi
        self.allowable_agent_type_list = []
        self.allowed_uses = []

        self.is_reversed = False
        self.is_cycle = False
        self.is_pure_cycle = False          # cycle without crossing nodes
        self.ref_node_id_list = []
        self.ref_node_list = []
        self.number_of_segments = 0
        self.segment_node_list = []         # ref node sequence for each segment

    def getNodeListForSegments(self):
        number_of_ref_nodes = len(self.ref_node_list)
        last_idx = 0
        idx = 0

        while True:
            m_segment_node_list = [self.ref_node_list[last_idx]]
            for idx in range(last_idx+1,number_of_ref_nodes):
                ref_node = self.ref_node_list[idx]
                m_segment_node_list.append(ref_node)
                if ref_node.is_crossing:
                    last_idx = idx
                    break
            self.segment_node_list.append(m_segment_node_list)
            self.number_of_segments += 1
            if idx == number_of_ref_nodes-1: break


class Relation:
    def __init__(self):
        self.osm_relation_id = None
        self.member_id_list = []
        self.member_type_list = []
        self.member_list = []
        self.member_role_list = []
        self.name = ''
        self.building = None
        self.amenity = None
        self.leisure = None


class OSMNetwork:
    def __init__(self):
        self.osm_node_dict = {}
        self.osm_way_dict = {}
        self.osm_relation_list = []

        self.link_way_list = []
        self.POI_way_list = []

        self.bounds = None
        self.GT = None




from osm2gmns.networkclass.macronet import Link
import osm2gmns.settings as og_settings
from shapely import ops


def _checkLinkAttr(ib_link, ob_link):
    if ib_link.name != ob_link.name: return False
    if ib_link.link_type != ob_link.link_type: return False
    if ib_link.is_link != ob_link.is_link: return False
    if ib_link.free_speed != ob_link.free_speed: return False
    if set(ib_link.allowed_uses) != set(ob_link.allowed_uses): return False

    if len(ib_link.lanes_list) > 1 or len(ob_link.lanes_list) > 1: return False
    if ib_link.lanes != ob_link.lanes: return False

    if ib_link.from_bidirectional_way != ob_link.from_bidirectional_way: return False
    if ib_link.ctrl_type != ob_link.ctrl_type: return False
    return True


def _newLinkFromLinks(link_id, up_link, down_link):
    link = Link(link_id)
    link.osm_way_id = f'{up_link.osm_way_id};{down_link.osm_way_id}'
    link.name = up_link.name
    link.link_class = up_link.link_class
    link.link_type_name = up_link.link_type_name
    link.link_type = up_link.link_type
    link.is_link = up_link.is_link
    link.free_speed = up_link.free_speed
    link.capacity = up_link.capacity
    link.allowed_uses = up_link.allowed_uses
    link.from_bidirectional_way = up_link.from_bidirectional_way
    link.lanes_list = up_link.lanes_list
    link.ctrl_type = up_link.ctrl_type

    link.from_node = up_link.from_node
    link.to_node = down_link.to_node
    link.from_node.outgoing_link_list.remove(up_link)
    link.to_node.incoming_link_list.remove(down_link)
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)

    link.geometry = ops.linemerge([up_link.geometry, down_link.geometry])
    link.geometry_xy = ops.linemerge([up_link.geometry_xy, down_link.geometry_xy])
    return link


def _combLinks(network):

    removal_node_set = set()
    removal_link_set = set()

    for node_id, node in network.node_dict.items():
        if node.osm_highway and 'traffic_signals' in node.osm_highway: continue

        if len(node.incoming_link_list) != 1 or len(node.outgoing_link_list) != 1:
            continue

        ib_link, ob_link = node.incoming_link_list[0], node.outgoing_link_list[0]
        if ib_link.from_node is ob_link.to_node:
            continue

        if (ib_link.ctrl_type is not None) or (ob_link.ctrl_type is not None):
            continue

        if not _checkLinkAttr(ib_link, ob_link):
            continue

        new_link = _newLinkFromLinks(network.max_link_id, ib_link, ob_link)
        network.link_dict[new_link.link_id] = new_link
        network.max_link_id += 1

        removal_node_set.add(node.node_id)
        removal_link_set.add(ib_link.link_id)
        removal_link_set.add(ob_link.link_id)

    for node_id in removal_node_set: del network.node_dict[node_id]
    for link_id in removal_link_set: del network.link_dict[link_id]



def combineShortLinks(network):

    if og_settings.verbose:
        print('    combining links')

    number_of_nodes_before_combination = len(network.node_dict)
    number_of_links_before_combination = len(network.link_dict)

    _combLinks(network)

    if og_settings.verbose:
        print(f'    link combination finished. before: {number_of_nodes_before_combination} nodes {number_of_links_before_combination} links, after {len(network.node_dict)} nodes {len(network.link_dict)} links')



import math
from shapely import geometry

_indent = 8.0

def getLinkAngle(ib_link, ob_link):
    angle_ib = math.atan2(ib_link.geometry_xy.coords[-1][1] - ib_link.geometry_xy.coords[0][1],
                          ib_link.geometry_xy.coords[-1][0] - ib_link.geometry_xy.coords[0][0])

    angle_ob = math.atan2(ob_link.geometry_xy.coords[-1][1] - ob_link.geometry_xy.coords[0][1],
                          ob_link.geometry_xy.coords[-1][0] - ob_link.geometry_xy.coords[0][0])

    angle = angle_ob - angle_ib
    if angle < -1 * math.pi:
        angle += 2 * math.pi
    if angle > math.pi:
        angle -= 2 * math.pi
    return angle


def getMovementDescription(ib_link, ob_link):
    ib_start, ib_end = ib_link.geometry_xy.coords[0], ib_link.geometry_xy.coords[-1]
    ob_end = ob_link.geometry_xy.coords[-1]

    angle_ib = math.atan2(ib_end[1] - ib_start[1], ib_end[0] - ib_start[0])
    if -0.75 * math.pi <= angle_ib < -0.25 * math.pi:
        direction = 'SB'
    elif -0.25 * math.pi <= angle_ib < 0.25 * math.pi:
        direction = 'EB'
    elif 0.25 * math.pi <= angle_ib < 0.75 * math.pi:
        direction = 'NB'
    else:
        direction = 'WB'

    angle_ob = math.atan2(ob_end[1] - ib_end[1], ob_end[0] - ib_end[0])
    angle = angle_ob - angle_ib
    if angle < -1 * math.pi:
        angle += 2 * math.pi
    if angle > math.pi:
        angle -= 2 * math.pi

    if -0.25 * math.pi <= angle <= 0.25 * math.pi:
        mvmt = 'T'
        mvmt_type = 'thru'
    elif angle < -0.25 * math.pi:
        mvmt = 'R'
        mvmt_type = 'right'
    elif angle <= 0.75 * math.pi:
        mvmt = 'L'
        mvmt_type = 'left'
    else:
        mvmt = 'U'
        mvmt_type = 'uturn'

    mvmt_txt_id = direction + mvmt
    return mvmt_txt_id, mvmt_type


def getMovementGeometry(ib_link, ob_link):
    ib_geometry_xy = ib_link.geometry_xy
    ib_indent = _indent if ib_geometry_xy.length > _indent else ib_geometry_xy.length / 2
    ib_point = ib_geometry_xy.interpolate(-1 * ib_indent)

    ob_geometry_xy = ob_link.geometry_xy
    ob_indent = _indent if ob_geometry_xy.length > _indent else ob_geometry_xy.length / 2
    ob_point = ob_geometry_xy.interpolate(ob_indent)

    geometry_xy = geometry.LineString([ib_point, ob_point])
    return geometry_xy





from osm2gmns.utils.util_coord import to_latlon, from_latlon
import osm2gmns.settings as og_settings
from shapely import geometry
import functools
import numpy as np



def getLineFromNodes(node_list):
    if len(node_list) < 2: return None, None
    point_list = [node.geometry for node in node_list]
    line = geometry.LineString(point_list)
    point_list_xy = [node.geometry_xy for node in node_list]
    line_xy = geometry.LineString(point_list_xy)
    return line, line_xy


def getPolygonFromNodes(node_list):
    if len(node_list) < 3: return None, None
    point_list = [node.geometry for node in node_list]
    poly = geometry.Polygon(point_list)
    point_list_xy = [node.geometry_xy for node in node_list]
    poly_xy = geometry.Polygon(point_list_xy)
    return poly, poly_xy


def offsetLine(line, distance):
    coords = list(line.coords)
    offset_coord_list_temp = []
    for i in range(len(coords) - 1):
        start_x, start_y = coords[i]
        end_x, end_y = coords[i+1]

        delta_x = end_x - start_x
        delta_y = end_y - start_y
        length = (delta_x ** 2 + delta_y ** 2) ** 0.5
        offset_x = delta_y / length * distance
        offset_y = -1 * delta_x / length * distance

        offset_coord_list_temp.append(((start_x + offset_x, start_y + offset_y), (end_x + offset_x, end_y + offset_y)))

    coords_offset = [offset_coord_list_temp[0][0]]
    for i in range(len(offset_coord_list_temp) - 1):
        uf = offset_coord_list_temp[i][0]
        ut = offset_coord_list_temp[i][1]
        df = offset_coord_list_temp[i + 1][0]
        dt = offset_coord_list_temp[i + 1][1]

        d_utdf = ((ut[0] - df[0]) ** 2 + ((ut[1] - df[1]) ** 2)) ** 0.5
        if d_utdf < 0.1:
            x, y = (ut[0] + df[0]) * 0.5, (ut[1] + df[1]) * 0.5
            coords_offset.append((x, y))
        else:
            A = [[ut[1] - uf[1], uf[0] - ut[0]], [dt[1] - df[1], df[0] - dt[0]]]
            b = [(ut[1] - uf[1]) * uf[0] - (ut[0] - uf[0]) * uf[1],
                 (dt[1] - df[1]) * df[0] - (dt[0] - df[0]) * df[1]]
            A_mat = np.mat(A)
            b_mat = np.mat(b).T
            solution = np.linalg.inv(A_mat) * b_mat
            x, y = solution[0, 0], solution[1, 0]

            d_mut = ((ut[0] - x) ** 2 + ((ut[1] - y) ** 2)) ** 0.5
            if d_mut > 200:
                x, y = (ut[0] + df[0]) * 0.5, (ut[1] + df[1]) * 0.5
            coords_offset.append((x, y))

    coords_offset.append(offset_coord_list_temp[-1][1])
    return geometry.LineString(coords_offset)


class GeoTransformer:
    def __init__(self, central_lon, central_lat, northern):
        self.central_lon = central_lon
        self.central_lat = central_lat
        self.northern = northern

        self.from_latlon = functools.partial(from_latlon, central_longitude=self.central_lon)
        self.to_latlon = functools.partial(to_latlon, central_longitude=self.central_lon, northern=self.northern)

    def _from_latlon_(self, p):
        return np.round(self.from_latlon(*p), og_settings.local_coord_precision)

    def _to_latlon_(self, p):
        return np.round(self.to_latlon(*p), og_settings.lonlat_coord_precision)

    def _transform(self, shape, func):
        construct = shape.__class__

        if shape.type.startswith('Multi'):
            parts = [self._transform(geom, func) for geom in shape.geoms]
            return construct(parts)

        if shape.type == 'Point':
            return construct(list(map(func, shape.coords))[0])

        if shape.type in ('Point', 'LineString'):
            return construct(map(func, shape.coords))

        if shape.type == 'Polygon':
            exterior = map(func, shape.exterior.coords)
            rings = [map(func, ring.coords) for ring in shape.interiors]
            return construct(exterior, rings)


    def geo_from_latlon(self, shape):
        return self._transform(shape, self._from_latlon_)

    def geo_to_latlon(self, shape):
        return self._transform(shape, self._to_latlon_)



from osm2gmns.io.util_io import getFileHandle
import osm2gmns.settings as og_settings
from shapely import wkt
import os
import csv


def _outputMesoNodes(mesonet, mesonode_filepath, projection, encoding):
    outfile = getFileHandle(mesonode_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['node_id', 'zone_id', 'x_coord', 'y_coord', 'macro_node_id', 'macro_link_id', 'activity_type', 'is_boundary'])
    for mesonode_id, mesonode in mesonet.node_dict.items():
        if projection:
            x_coord = round(mesonode.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(mesonode.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(mesonode.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(mesonode.geometry.y, og_settings.lonlat_coord_precision)

        line = [mesonode.node_id, mesonode.zone_id, x_coord, y_coord, mesonode.macro_node_id, mesonode.macro_link_id,
                mesonode.activity_type, mesonode.is_boundary]
        writer.writerow(line)

    outfile.close()


def _outputMesoLinks(mesonet, mesolink_filepath, projection, encoding):
    outfile = getFileHandle(mesolink_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['link_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes', 'capacity',
                     'free_speed', 'link_type_name', 'link_type', 'geometry', 'macro_node_id', 'macro_link_id',
                     'mvmt_txt_id', 'allowed_uses'])
    for mesolink_id, mesolink in mesonet.link_dict.items():
        if projection:
            geometry_ = wkt.dumps(mesolink.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(mesolink.geometry, rounding_precision=og_settings.lonlat_coord_precision)

        line = [mesolink.link_id, mesolink.from_node.node_id, mesolink.to_node.node_id, mesolink.dir_flag, mesolink.length,
                mesolink.lanes, mesolink.capacity, mesolink.free_speed, mesolink.link_type_name,mesolink.link_type,
                geometry_, mesolink.macro_node_id, mesolink.macro_link_id, mesolink.mvmt_txt_id, ';'.join(mesolink.allowed_uses)]
        writer.writerow(line)

    outfile.close()


def _outputMicroNodes(micronet, micronode_filepath, projection, encoding):
    outfile = getFileHandle(micronode_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['node_id', 'zone_id', 'x_coord', 'y_coord', 'meso_link_id', 'lane_no', 'is_boundary'])
    for micronode_id, micronode in micronet.node_dict.items():
        if projection:
            x_coord = round(micronode.geometry_xy.x, og_settings.local_coord_precision)
            y_coord = round(micronode.geometry_xy.y, og_settings.local_coord_precision)
        else:
            x_coord = round(micronode.geometry.x, og_settings.lonlat_coord_precision)
            y_coord = round(micronode.geometry.y, og_settings.lonlat_coord_precision)

        line = [micronode.node_id, micronode.zone_id, x_coord, y_coord, micronode.mesolink.link_id, micronode.lane_no, micronode.is_boundary]
        writer.writerow(line)

    outfile.close()


def _outputMicroLinks(micronet, microlink_filepath, projection, encoding):
    outfile = getFileHandle(microlink_filepath, encoding)
    writer = csv.writer(outfile)

    writer.writerow(['link_id', 'from_node_id', 'to_node_id', 'dir_flag', 'length', 'lanes', 'capacity',
                     'free_speed', 'link_type_name', 'link_type', 'geometry', 'macro_node_id', 'macro_link_id',
                     'meso_link_id', 'cell_type', 'additional_cost', 'lane_no', 'mvmt_txt_id', 'allowed_uses'])
    for microlink_id, microlink in micronet.link_dict.items():
        if projection:
            geometry_ = wkt.dumps(microlink.geometry_xy, rounding_precision=og_settings.local_coord_precision)
        else:
            geometry_ = wkt.dumps(microlink.geometry, rounding_precision=og_settings.lonlat_coord_precision)

        line = [microlink.link_id, microlink.from_node.node_id, microlink.to_node.node_id, microlink.dir_flag,
                microlink.length, 1, microlink.mesolink.capacity, microlink.mesolink.free_speed,
                microlink.mesolink.link_type_name, microlink.mesolink.link_type, geometry_,
                microlink.mesolink.macro_node_id, microlink.mesolink.macro_link_id, microlink.mesolink.link_id,
                microlink.cell_type, 0, microlink.lane_no, microlink.mvmt_txt_id, ';'.join(microlink.allowed_uses)]
        writer.writerow(line)

    outfile.close()



def outputMesoNet(mesonet, output_folder, prefix, projection, encoding):
    mesonet_folder = os.path.join(output_folder, 'mesonet')
    if not os.path.isdir(mesonet_folder): os.mkdir(mesonet_folder)

    mesonode_filepath = os.path.join(mesonet_folder, f'{prefix}node.csv')
    _outputMesoNodes(mesonet, mesonode_filepath, projection, encoding)

    mesolink_filepath = os.path.join(mesonet_folder, f'{prefix}link.csv')
    _outputMesoLinks(mesonet, mesolink_filepath, projection, encoding)


def outputMicroNet(micronet, output_folder, prefix, projection, encoding):
    micronet_folder = os.path.join(output_folder, 'micronet')
    if not os.path.isdir(micronet_folder): os.mkdir(micronet_folder)

    micronode_filepath = os.path.join(micronet_folder, f'{prefix}node.csv')
    _outputMicroNodes(micronet, micronode_filepath, projection, encoding)

    microlink_filepath = os.path.join(micronet_folder, f'{prefix}link.csv')
    _outputMicroLinks(micronet, microlink_filepath, projection, encoding)



# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 15:34
# @desc         [script description]

import osm2gmns.settings as og_settings


class BaseNode:
    def __init__(self, node_id):
        self.name = None
        self.node_id = node_id
        self.geometry = None
        self.geometry_xy = None

        self.other_attrs = {}

        self.incoming_link_list = []
        self.outgoing_link_list = []



class BaseLink:
    def __init__(self, link_id):
        self.name = None
        self.link_id = link_id
        self.from_node = None
        self.to_node = None
        self.dir_flag = 1
        self.geometry = None
        self.geometry_xy = None

        self.other_attrs = {}

    @property
    def length(self):
        return round(self.geometry_xy.length, og_settings.local_coord_precision)



class BaseNetwork:
    def __init__(self):
        self.node_dict = {}
        self.link_dict = {}

        self.max_node_id = 0
        self.max_link_id = 0

        self.original_coordinate_type = 'lonlat'
        self.GT = None
        self.bounds = None

        self.node_other_attrs = []
        self.link_other_attrs = []

    @property
    def number_of_nodes(self):
        return len(self.node_dict)

    @property
    def number_of_links(self):
        return len(self.link_dict)




from osm2gmns.movement.autoconintd import CAutoConnectorIntD
from osm2gmns.movement.autoconm import CAutoConnectorM
from osm2gmns.movement.util_mvmt import getMovementDescription, getMovementGeometry
from osm2gmns.networkclass.macronet import Movement
import osm2gmns.settings as og_settings


def generateMovementsForOneNode(node, max_movement_id, GT):
    number_of_incoming_links = len(node.incoming_link_list)
    number_of_outgoing_links = len(node.outgoing_link_list)

    if number_of_incoming_links == 0 or number_of_outgoing_links == 0: return

    movement_id = max_movement_id

    if number_of_outgoing_links == 1:
        # merge
        ob_link = node.outgoing_link_list[0]
        ib_link_list = []
        for ib_link in node.incoming_link_list:
            if ib_link.from_node.node_id != ob_link.to_node.node_id:
                ib_link_list.append(ib_link)

        if len(ib_link_list) == 0: return

        CAutoConnectorM.ob_link, CAutoConnectorM.ib_link_list = ob_link, ib_link_list
        connection_list = CAutoConnectorM.buildConnector()

        for ib_link_no, ib_link in enumerate(ib_link_list):
            ib_lane_index_start = connection_list[ib_link_no][0][0] + 1
            ib_lane_index_end = connection_list[ib_link_no][0][1] + 1
            ob_lane_index_start = connection_list[ib_link_no][1][0] + 1
            ob_lane_index_end = connection_list[ib_link_no][1][1] + 1
            number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

            mvmt = Movement(movement_id)
            mvmt.node = node
            mvmt.ib_link = ib_link
            mvmt.ob_link = ob_link
            mvmt.start_ib_lane, mvmt.end_ib_lane = ib_lane_index_start, ib_lane_index_end
            mvmt.start_ob_lane, mvmt.end_ob_lane = ob_lane_index_start, ob_lane_index_end
            mvmt.lanes = number_of_lanes
            mvmt_txt_id, mvmt_type = getMovementDescription(ib_link, ob_link)
            mvmt.mvmt_txt_id = mvmt_txt_id
            mvmt.type = mvmt_type
            mvmt.geometry_xy = getMovementGeometry(ib_link, ob_link)
            mvmt.geometry = GT.geo_to_latlon(mvmt.geometry_xy)
            mvmt.generated_by_osm2gmns = True

            if node.ctrl_type == 1: mvmt.ctrl_type = 'signal'               # todo: check ctrl_type
            mvmt.allowed_uses = ib_link.allowed_uses
            node.movement_list.append(mvmt)
            movement_id += 1
    else:
        # diverge and intersections
        for ib_link in node.incoming_link_list:
            ob_link_list = []
            for ob_link in node.outgoing_link_list:
                if ib_link.from_node.node_id != ob_link.to_node.node_id:
                    ob_link_list.append(ob_link)

            if len(ob_link_list) == 0: continue

            CAutoConnectorIntD.ib_link, CAutoConnectorIntD.ob_link_list = ib_link, ob_link_list
            connection_list = CAutoConnectorIntD.buildConnector()

            for ob_link_no, ob_link in enumerate(ob_link_list):
                ib_lane_index_start = connection_list[ob_link_no][0][0] + 1
                ib_lane_index_end = connection_list[ob_link_no][0][1] + 1
                ob_lane_index_start = connection_list[ob_link_no][1][0] + 1
                ob_lane_index_end = connection_list[ob_link_no][1][1] + 1
                number_of_lanes = ib_lane_index_end - ib_lane_index_start + 1

                mvmt = Movement(movement_id)
                mvmt.node = node
                mvmt.ib_link = ib_link
                mvmt.ob_link = ob_link
                mvmt.start_ib_lane, mvmt.end_ib_lane = ib_lane_index_start, ib_lane_index_end
                mvmt.start_ob_lane, mvmt.end_ob_lane = ob_lane_index_start, ob_lane_index_end
                mvmt.lanes = number_of_lanes
                mvmt_txt_id, mvmt_type = getMovementDescription(ib_link, ob_link)
                mvmt.mvmt_txt_id = mvmt_txt_id
                mvmt.type = mvmt_type
                mvmt.geometry_xy = getMovementGeometry(ib_link, ob_link)
                mvmt.geometry = GT.geo_to_latlon(mvmt.geometry_xy)
                mvmt.generated_by_osm2gmns = True

                if node.ctrl_type == 1: mvmt.ctrl_type = 'signal'
                mvmt.allowed_uses = ib_link.allowed_uses
                node.movement_list.append(mvmt)
                movement_id += 1


def generateMovements(network):
    """
    Use osm2gmns built-in methods to generate movements for each node (intersection) in a network

    Parameters
    ----------
    network: Network
        an osm2gmns Network object

    Returns
    -------
    None
    """

    if og_settings.verbose:
        print('Generating Movements')

    max_movement_id = network.max_movement_id
    first_intersection_lacking_lanes = True

    for node_id, node in network.node_dict.items():

        if node.movement_list:
            continue

        lack_lanes = False
        for link in node.outgoing_link_list + node.incoming_link_list:
            if link.lanes is None:
                lack_lanes = True
                break

        if lack_lanes:
            if first_intersection_lacking_lanes:
                print('WARNING: Movement generation will be skipped at some intersections due to the lack of link lanes information.\n'
                      '         Set "default_lanes" as "True" when parsing networks to assign default lanes for links without lanes information.')
                first_intersection_lacking_lanes = False
            continue

        generateMovementsForOneNode(node, max_movement_id, network.GT)
        max_movement_id += len(node.movement_list)

    network.max_movement_id = max_movement_id



